from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Sequence,
    TypeVar,
)

from whitecanvas.canvas.dataframe._base import BaseCatPlotter
from whitecanvas.layers import tabular as _lt
from whitecanvas.layers.tabular import _jitter
from whitecanvas.types import HistBinType

if TYPE_CHECKING:
    from whitecanvas.canvas import JointGrid

    NStr = str | Sequence[str]

_C = TypeVar("_C", bound="JointGrid")
_DF = TypeVar("_DF")


class JointCatPlotter(BaseCatPlotter[_C, _DF]):
    def __init__(
        self,
        canvas: _C,
        df: _DF,
        x: str | None,
        y: str | None,
        update_labels: bool = False,
    ):
        super().__init__(canvas, df)
        self._x = x
        self._y = y
        self._update_label = update_labels
        if update_labels:
            self._update_xy_label(x, y)

    def _get_x(self) -> str:
        if self._x is None:
            raise ValueError("Column for x-axis is not set")
        return self._x

    def _get_y(self) -> str:
        if self._y is None:
            raise ValueError("Column for y-axis is not set")
        return self._y

    def _update_xy_label(self, x: str | None, y: str | None) -> None:
        """Update the x and y labels using the column names"""
        canvas = self._canvas()
        if isinstance(x, str):
            canvas.x.label.text = x
        if isinstance(y, str):
            canvas.y.label.text = y

    def add_markers(
        self,
        *,
        name: str | None = None,
        color: NStr | None = None,
        hatch: NStr | None = None,
        size: str | None = None,
        symbol: NStr | None = None,
    ) -> _lt.DFMarkers[_DF]:
        """
        Add a categorical marker plot.

        Parameters
        ----------
        name : str, optional
            Name of the layer.
        color : str or sequence of str, optional
            Column name(s) for coloring the lines. Must be categorical.
        hatch : str or sequence of str, optional
            Column name(s) for hatches. Must be categorical.
        size : str, optional
            Column name for marker size. Must be numerical.
        symbol : str or sequence of str, optional
            Column name(s) for symbols. Must be categorical.

        Returns
        -------
        DFMarkers
            Marker collection layer.
        """
        grid = self._canvas()
        main = grid.main_canvas
        xj = _jitter.IdentityJitter(self._get_x())
        yj = _jitter.IdentityJitter(self._get_y())
        layer = _lt.DFMarkers(
            self._df, xj, yj, name=name, color=color, hatch=hatch,
            size=size, symbol=symbol, backend=grid._backend,
        )  # fmt: skip
        if color is not None and not layer._color_by.is_const():
            layer.update_color(layer._color_by.by, palette=main._color_palette)
        elif color is None:
            layer.update_color(main._color_palette.next())
        main.add_layer(layer)
        for _x_plt in grid._iter_x_plotters():
            xlayer = _x_plt.add_layer_for_cat_markers(
                self._df, self._get_x(), color=color, hatch=hatch, backend=grid._backend
            )  # fmt: skip
            grid.x_canvas.add_layer(xlayer)
            grid._link_marginal_to_main(xlayer, layer)
        for _y_plt in grid._iter_y_plotters():
            ylayer = _y_plt.add_layer_for_cat_markers(
                self._df, self._get_y(), color=color, hatch=hatch, backend=grid._backend
            )
            grid.y_canvas.add_layer(ylayer)
            grid._link_marginal_to_main(ylayer, layer)
        grid._autoscale_layers()
        return layer

    def add_hist2d(
        self,
        *,
        color: str | Sequence[str] | None = None,
        name: str | None = None,
        bins: HistBinType | tuple[HistBinType, HistBinType] = "auto",
        rangex: tuple[float, float] | None = None,
        rangey: tuple[float, float] | None = None,
    ) -> _lt.DFHeatmap[_DF]:
        """
        Add 2-D histogram of given x/y columns.

        Parameters
        ----------
        color : str or sequence of str, optional
            Column name(s) for coloring the heatmaps. Must be categorical.
        name : str, optional
            Name of the layer.
        bins : int, array, str or tuple of them, default "auto"
            If int, the number of bins for both x and y. If tuple, the number of bins
            for x and y respectively.
        rangex : (float, float), optional
            Range of x values in which histogram will be built.
        rangey : (float, float), optional
            Range of y values in which histogram will be built.

        Returns
        -------
        DFHeatmap
            Dataframe bound heatmap layer.
        """
        grid = self._canvas()
        main = grid.main_canvas
        layer = _lt.DFMultiHeatmap.build_hist(
            self._df, self._get_x(), self._get_y(), color=color, name=name, bins=bins,
            range=(rangex, rangey), backend=grid._backend,
        )  # fmt: skip
        main.add_layer(layer)
        for _x_plt in grid._iter_x_plotters():
            xlayer = _x_plt.add_layer_for_cat_hist2d(
                self._df, self._get_x(), bins=bins, limits=rangex, backend=grid._backend
            )  # fmt: skip
            grid.x_canvas.add_layer(xlayer)
            grid._link_marginal_to_main(xlayer, layer)
        for _y_plt in grid._iter_y_plotters():
            ylayer = _y_plt.add_layer_for_cat_hist2d(
                self._df, self._get_y(), bins=bins, limits=rangey, backend=grid._backend
            )  # fmt: skip
            grid.y_canvas.add_layer(ylayer)
            grid._link_marginal_to_main(ylayer, layer)
        grid._autoscale_layers()
        return layer



from __future__ import annotations

import sys
import warnings
import weakref
from typing import TYPE_CHECKING, Callable, cast

import numpy as np
from numpy.typing import NDArray
from psygnal import Signal
from vispy import use as vispy_use
from vispy.scene import PanZoomCamera, SceneCanvas, ViewBox, visuals
from vispy.util import keys

from whitecanvas import protocols
from whitecanvas.backend.vispy._gridlines import GridLines
from whitecanvas.backend.vispy._label import Axis, TextLabel, Ticks
from whitecanvas.types import Modifier, MouseButton, MouseEvent, MouseEventType

if TYPE_CHECKING:
    from vispy.app.canvas import MouseEvent as vispyMouseEvent
    from vispy.scene import Grid
    from vispy.scene.subscene import SubScene
    from vispy.visuals import Visual


class Camera(PanZoomCamera):
    changed = Signal()

    def view_changed(self):
        super().view_changed()
        self.changed.emit()


@protocols.check_protocol(protocols.CanvasProtocol)
class Canvas:
    def __init__(self, viewbox: ViewBox):
        self._outer_viewbox = viewbox
        grid = cast("Grid", viewbox.add_grid())
        grid.spacing = 0
        _viewbox = grid.add_view(row=1, col=1, camera=Camera())
        self._viewbox: ViewBox = _viewbox

        title = TextLabel("")
        title.height_max = 40
        grid.add_widget(title, row=0, col=0, col_span=2)
        self._title = title
        x_axis = Axis(
            self,
            dim=1,
            orientation="bottom",
            anchors=("center", "bottom"),
            font_size=6,
            axis_label_margin=40,
            tick_label_margin=5,
            axis_label="",
        )
        x_axis.stretch = (1, 0.1)
        x_axis.height_min = x_axis.height_max = 40
        grid.add_widget(x_axis, row=2, col=1)
        x_axis.link_view(self._viewbox)
        y_axis = Axis(
            self,
            dim=0,
            orientation="left",
            anchors=("right", "middle"),
            font_size=6,
            axis_label_margin=50,
            tick_label_margin=5,
            axis_label="",
        )

        y_axis.stretch = (0.1, 1)
        grid.add_widget(y_axis, row=1, col=0)
        y_axis.link_view(self._viewbox)
        self._grid_lines = GridLines(self)
        self._xaxis = x_axis
        self._yaxis = y_axis
        self._xticks = Ticks(x_axis)
        self._yticks = Ticks(y_axis)
        self._xlabel = TextLabel("")
        self._ylabel = TextLabel("")
        self._grid = grid
        self._mouse_click_callbacks: list[Callable[[MouseEvent], None]] = []
        self._mouse_move_callbacks: list[Callable[[MouseEvent], None]] = []
        self._mouse_double_click_callbacks: list[Callable[[MouseEvent], None]] = []
        self._mouse_release_callbacks: list[Callable[[MouseEvent], None]] = []

    def _set_scene_ref(self, scene):
        self._viewbox.unfreeze()
        self._viewbox._canvas_ref = weakref.ref(self)
        self._viewbox._scene_ref = weakref.ref(scene)
        self._viewbox.freeze()

    def _plt_get_native(self):
        return self._outer_viewbox

    def _plt_get_title(self):
        return self._title

    def _plt_get_xaxis(self):
        return self._xaxis

    def _plt_get_yaxis(self):
        return self._yaxis

    def _plt_get_xlabel(self):
        return self._xlabel

    def _plt_get_xticks(self):
        return self._xticks

    def _plt_get_yticks(self):
        return self._yticks

    def _plt_get_ylabel(self):
        return self._ylabel

    def _plt_reorder_layers(self, layers: list[protocols.BaseProtocol]):
        """Reorder layers in the canvas"""
        vb = self._viewbox
        for idx, layer in enumerate(layers):
            layer.order = idx
        if hasattr(vb, "_scene_ref"):
            scene: SceneCanvas = vb._scene_ref()
            scene._draw_order.clear()
            scene.update()

    @property
    def _camera(self) -> Camera:
        return self._viewbox.camera

    def _plt_get_aspect_ratio(self) -> float | None:
        """Get aspect ratio of canvas"""
        return self._camera.aspect

    def _plt_set_aspect_ratio(self, ratio: float | None):
        """Set aspect ratio of canvas"""
        self._camera.aspect = ratio

    def _plt_add_layer(self, layer: visuals.visuals.Visual):
        layer.set_gl_state(
            depth_test=False,
            cull_face=False,
            blend=True,
            blend_func=("src_alpha", "one_minus_src_alpha", "one", "one"),
            blend_equation="func_add",
        )
        layer.parent = self._viewbox.scene

    def _plt_remove_layer(self, layer):
        """Remove layer from the canvas"""
        layer.parent = None

    def _plt_get_visible(self) -> bool:
        """Get visibility of canvas"""
        return self._grid.visible

    def _plt_set_visible(self, visible: bool):
        """Set visibility of canvas"""
        self._grid.visible = visible

    @property
    def _scene(self) -> SceneCanvas:
        return self._viewbox.scene

    def _plt_connect_mouse_click(self, callback: Callable[[MouseEvent], None]):
        """Connect callback to clicked event"""
        self._mouse_click_callbacks.append(callback)

    def _plt_connect_mouse_drag(self, callback: Callable[[MouseEvent], None]):
        """Connect callback to clicked event"""
        self._mouse_move_callbacks.append(callback)

    def _plt_connect_mouse_double_click(self, callback: Callable[[MouseEvent], None]):
        """Connect callback to clicked event"""
        self._mouse_double_click_callbacks.append(callback)

    def _plt_connect_mouse_release(self, callback: Callable[[MouseEvent], None]):
        """Connect callback to clicked event"""
        self._mouse_release_callbacks.append(callback)

    def _plt_connect_xlim_changed(
        self, callback: Callable[[tuple[float, float]], None]
    ):
        @self._camera.changed.connect
        def _cb():
            with self._camera.changed.blocked():
                callback(self._xaxis._plt_get_limits())

    def _plt_connect_ylim_changed(
        self, callback: Callable[[tuple[float, float]], None]
    ):
        @self._camera.changed.connect
        def _cb():
            with self._camera.changed.blocked():
                callback(self._yaxis._plt_get_limits())

    def _plt_draw(self):
        pass  # vispy has its own draw mechanism

    def _plt_make_legend(self, *args, **kwargs):
        warnings.warn(
            "Legend is not supported in vispy backend",
            UserWarning,
            stacklevel=2,
        )

    def _plt_get_mouse_enabled(self) -> bool:
        return self._camera.interactive

    def _plt_set_mouse_enabled(self, enabled: bool):
        self._camera.interactive = enabled


@protocols.check_protocol(protocols.CanvasGridProtocol)
class CanvasGrid:
    def __init__(self, heights: list[float], widths: list[float], app: str = "default"):
        if app == "qt":  # pragma: no cover
            import importlib

            for mod in ["PyQt5", "PyQt6", "PyQt4"]:
                try:
                    importlib.import_module(mod)
                except ImportError:
                    pass
                else:
                    app = mod.lower()
                    break
            else:
                raise ValueError("No Qt bindings found.")
        if app != "default":
            vispy_use(_APP_NAMES.get(app, app))
        self._scene = SceneCanvasExt(keys="interactive")
        self._grid: Grid = self._scene.central_widget.add_grid()
        self._scene.create_native()
        self._heights = heights
        self._widths = widths

    def _calc_rc_and_span(self, row: int, col: int, rowspan: int, colspan: int):
        rspan = sum(self._heights[row : row + rowspan])
        cspan = sum(self._widths[col : col + colspan])
        r = sum(self._heights[:row])
        c = sum(self._widths[:col])
        return r, c, rspan, cspan

    def _plt_add_canvas(self, row: int, col: int, rowspan: int, colspan: int):
        view_pos = self._calc_rc_and_span(row, col, rowspan, colspan)
        viewbox: ViewBox = self._grid.add_view(*view_pos)
        canvas = Canvas(viewbox)
        canvas._set_scene_ref(self._scene)
        return canvas

    def _plt_add_canvas_3d(self, row: int, col: int, rowspan: int, colspan: int):
        from whitecanvas.backend.vispy.components3d import Canvas3D

        view_pos = self._calc_rc_and_span(row, col, rowspan, colspan)
        viewbox: ViewBox = self._grid.add_view(*view_pos)
        canvas = Canvas3D(viewbox)
        # canvas._set_scene_ref(self._scene)
        return canvas

    def _plt_get_background_color(self):
        return self._scene.bgcolor.rgba

    def _plt_set_background_color(self, color):
        self._scene.bgcolor = color

    def _plt_screenshot(self) -> NDArray[np.uint8]:
        return self._scene.render()

    def _plt_show(self):
        self._scene.show()

    def _plt_set_figsize(self, width: int, height: int):
        self._scene.size = (width, height)

    def _plt_set_spacings(self, wspace: float, hspace: float):
        self._grid.spacing = wspace


_APP_NAMES = {
    "qt4": "pyqt4",
    "qt5": "pyqt5",
    "qt6": "pyqt6",
    "tk": "tkinter",
    "notebook": "jupyter_rfb",
}


class SceneCanvasExt(SceneCanvas):
    scene: SubScene

    def on_mouse_press(self, event: vispyMouseEvent):
        visual = self.visual_at(event.pos)
        if isinstance(visual, ViewBox) and hasattr(visual, "_canvas_ref"):
            canvas: Canvas = visual._canvas_ref()
            tr = self.scene.node_transform(visual.scene)
            pos = tr.map(event.pos)[:2]
            ev = MouseEvent(
                button=_VISPY_BUTTON_MAP.get(event.button, MouseButton.NONE),
                modifiers=tuple(_VISPY_KEY_MAP[mod] for mod in event.modifiers),
                pos=pos,
                type=MouseEventType.PRESS,
            )
            for callback in canvas._mouse_click_callbacks:
                callback(ev)

    def on_mouse_move(self, event: vispyMouseEvent):
        visual = self.visual_at(event.pos)
        if isinstance(visual, ViewBox) and hasattr(visual, "_canvas_ref"):
            canvas: Canvas = visual._canvas_ref()
            tr = self.scene.node_transform(visual.scene)
            pos = tr.map(event.pos)[:2]
            ev = MouseEvent(
                button=_VISPY_BUTTON_MAP.get(event.button, MouseButton.NONE),
                modifiers=tuple(_VISPY_KEY_MAP[mod] for mod in event.modifiers),
                pos=pos,
                type=MouseEventType.MOVE,
            )

            for callback in canvas._mouse_move_callbacks:
                callback(ev)

    def on_mouse_double_click(self, event: vispyMouseEvent):
        visual = self.visual_at(event.pos)
        if isinstance(visual, ViewBox) and hasattr(visual, "_canvas_ref"):
            canvas: Canvas = visual._canvas_ref()
            tr = self.scene.node_transform(visual.scene)
            pos = tr.map(event.pos)[:2]
            ev = MouseEvent(
                button=_VISPY_BUTTON_MAP.get(event.button, MouseButton.NONE),
                modifiers=tuple(_VISPY_KEY_MAP[mod] for mod in event.modifiers),
                pos=pos,
                type=MouseEventType.DOUBLE_CLICK,
            )

            for callback in canvas._mouse_double_click_callbacks:
                callback(ev)

    def on_mouse_release(self, event: vispyMouseEvent):
        visual = self.visual_at(event.pos)
        if isinstance(visual, ViewBox) and hasattr(visual, "_canvas_ref"):
            canvas: Canvas = visual._canvas_ref()
            tr = self.scene.node_transform(visual.scene)
            pos = tr.map(event.pos)[:2] - 0.5
            ev = MouseEvent(
                button=_VISPY_BUTTON_MAP.get(event.button, MouseButton.NONE),
                modifiers=tuple(_VISPY_KEY_MAP[mod] for mod in event.modifiers),
                pos=pos,
                type=MouseEventType.RELEASE,
            )

            for callback in canvas._mouse_release_callbacks:
                callback(ev)


def as_overlay(layer: Visual, canvas: Canvas):
    layer.parent = canvas._outer_viewbox.scene
    layer.order = 10000


_VISPY_KEY_MAP = {
    keys.SHIFT: Modifier.SHIFT,
    keys.CONTROL: Modifier.CTRL,
    keys.ALT: Modifier.ALT,
    keys.META: Modifier.META,
}

_VISPY_BUTTON_MAP = {
    1: MouseButton.LEFT,
    2: MouseButton.RIGHT,
    3: MouseButton.MIDDLE,
    4: MouseButton.BACK,
    5: MouseButton.FORWARD,
}



from whitecanvas.backend.mock.canvas import Canvas, CanvasGrid
from whitecanvas.backend.mock.layers import (
    Band,
    Bars,
    Image,
    Markers,
    MonoLine,
    MultiLine,
    Texts,
    Vectors,
)



from __future__ import annotations

from functools import reduce

import numpy as np
import pyqtgraph as pg
from numpy.typing import NDArray
from pyqtgraph.GraphicsScene.mouseEvents import HoverEvent as pgHoverEvent
from qtpy import QtCore, QtGui

from whitecanvas.backend.pyqtgraph._base import PyQtLayer
from whitecanvas.backend.pyqtgraph._qt_utils import (
    array_to_qcolor,
    from_qt_line_style,
    to_qt_line_style,
)
from whitecanvas.protocols import LineProtocol, MultiLineProtocol, check_protocol
from whitecanvas.types import LineStyle
from whitecanvas.utils.type_check import is_real_number


@check_protocol(LineProtocol)
class MonoLine(pg.PlotCurveItem, PyQtLayer):
    def __init__(self, xdata, ydata):
        pen = QtGui.QPen(QtGui.QColor(0, 0, 0))
        pen.setCosmetic(True)
        super().__init__(xdata, ydata, pen=pen, antialias=False, clickable=True)
        self._hover_texts: list[str] | None = None
        self._toolTipCleared = True

    ##### XYDataProtocol #####
    def _plt_get_data(self):
        return self.getData()

    def _plt_set_data(self, xdata, ydata):
        self.setData(xdata, ydata)

    ##### HasEdges #####
    def _get_pen(self) -> QtGui.QPen:
        return self.opts["pen"]

    def _plt_get_edge_width(self) -> float:
        return self._get_pen().widthF()

    def _plt_set_edge_width(self, width: float):
        pen = self._get_pen()
        pen.setWidthF(width)
        self.setPen(pen)

    def _plt_get_edge_style(self) -> LineStyle:
        return from_qt_line_style(self._get_pen().style())

    def _plt_set_edge_style(self, style: LineStyle):
        pen = self._get_pen()
        pen.setStyle(to_qt_line_style(style))
        self.setPen(pen)

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        return np.array(self._get_pen().color().getRgbF())

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        pen = self._get_pen()
        pen.setColor(array_to_qcolor(color))
        self.setPen(pen)

    def _plt_get_antialias(self) -> bool:
        return self.opts["antialias"]

    def _plt_set_antialias(self, antialias: bool):
        self.opts["antialias"] = antialias
        self.update()

    def closestPoint(self, pos: QtCore.QPointF) -> int:
        # TODO: pick the closest point for now, but ideally, we should pick the
        # closest edge first, and then the closest point on the edge
        x, y = self.getData()
        dist2 = (x - pos.x()) ** 2 + (y - pos.y()) ** 2
        idx = np.argmin(dist2)
        if dist2[idx] < 10:
            return idx
        return idx

    def _plt_connect_pick_event(self, callback):
        def cb(ins, ev: QtGui.QMouseEvent):
            callback([self.closestPoint(ev.pos())])

        self.sigClicked.connect(cb)

    def _plt_set_hover_text(self, text: list[str]):
        self._hover_texts = text

    def hoverEvent(self, ev: pgHoverEvent):
        vb = self.getViewBox()
        if vb is not None and self._hover_texts is not None:
            if self.mouseShape().contains(ev.pos()):
                i = self.closestPoint(ev.pos())
                vb.setToolTip(self._hover_texts[i])
                self._toolTipCleared = False
            else:
                if not self._toolTipCleared:
                    vb.setToolTip("")
                    self._toolTipCleared = True


@check_protocol(MultiLineProtocol)
class MultiLine(pg.ItemGroup, PyQtLayer):
    clicked = QtCore.Signal(list)

    def __init__(self, data: list[NDArray[np.floating]]):
        super().__init__()
        self._lines: list[pg.PlotCurveItem] = []
        self._qpens: list[QtGui.QPen] = []
        for i, seg in enumerate(data):
            pen = QtGui.QPen(QtGui.QColor(255, 255, 255))
            pen.setCosmetic(True)
            item = pg.PlotCurveItem(
                seg[:, 0], seg[:, 1], pen=pen, antialias=True, clickable=True
            )
            item.sigClicked.connect(self._clicked_cb(i))
            self.addItem(item)
            self._lines.append(item)
            self._qpens.append(pen)
        self._data = data
        self._bounding_rect_cache = None
        self._antialias = True
        self._hover_texts: list[str] = None
        self._toolTipCleared = True

    def _clicked_cb(self, idx: int):
        return lambda _ins, _: self.clicked.emit([idx])

    def boundingRect(self):
        if self._bounding_rect_cache is not None:
            return self._bounding_rect_cache
        if len(self._lines) == 0:
            return QtCore.QRectF()
        rect = reduce(lambda a, b: a | b, (item.boundingRect() for item in self._lines))
        self._bounding_rect_cache = rect
        return rect

    ##### XYDataProtocol #####
    def _plt_get_data(self):
        return self._data

    def _plt_set_data(self, data: list[NDArray[np.floating]]):
        ndata = len(data)
        nitem = len(self._lines)
        scene = self.scene()
        if ndata < nitem:
            for item in self._lines[ndata:]:
                scene.removeItem(item)
                item.sigClicked.disconnect()
            self._lines = self._lines[:ndata]
            self._qpens = self._qpens[:ndata]
        else:
            # pick last pen
            if nitem > 0:
                pen = self._qpens[-1]
            else:
                pen = QtGui.QPen(QtGui.QColor(255, 255, 255))
            for i in range(nitem, ndata):
                item = pg.PlotCurveItem(pen=pen, antialias=True, clickable=True)
                item.sigClicked.connect(self._clicked_cb(i))
                self.addItem(item)
                self._lines.append(item)
                self._qpens.append(pen)
        for item, seg in zip(self._lines, data):
            item.setData(seg[:, 0], seg[:, 1])
        self._data = data
        self._bounding_rect_cache = None

    ##### HasEdges #####
    def _set_pen_to_curves(self):
        for item, pen in zip(self._lines, self._qpens):
            item.setPen(pen)

    def _plt_get_edge_width(self) -> NDArray[np.float32]:
        return np.array([pen.widthF() for pen in self._qpens], dtype=np.float32)

    def _plt_set_edge_width(self, width: float | NDArray[np.float32]):
        if is_real_number(width):
            widths: list[float] = [width] * len(self._qpens)
        else:
            widths = width
        for pen, w in zip(self._qpens, widths):
            pen.setWidthF(w)
        self._set_pen_to_curves()

    def _plt_get_edge_style(self) -> list[LineStyle]:
        return [from_qt_line_style(pen.style()) for pen in self._qpens]

    def _plt_set_edge_style(self, style: LineStyle | list[LineStyle]):
        if isinstance(style, LineStyle):
            styles: list[LineStyle] = [style] * len(self._qpens)
        else:
            styles = style
        for pen, s in zip(self._qpens, styles):
            pen.setStyle(to_qt_line_style(s))
        self._set_pen_to_curves()

    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        return np.array(
            [pen.color().getRgbF() for pen in self._qpens], dtype=np.float32
        )

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        if color.ndim == 1:
            colors = [color] * len(self._qpens)
        else:
            colors = color
        for pen, c in zip(self._qpens, colors):
            pen.setColor(array_to_qcolor(c))
        self._set_pen_to_curves()

    def _plt_get_antialias(self) -> bool:
        return self._antialias

    def _plt_set_antialias(self, antialias: bool):
        for item in self._lines:
            item.opts["antialias"] = antialias
            item.update()
        self._antialias = antialias

    def hoverEvent(self, ev: pgHoverEvent):
        vb = self.getViewBox()
        if vb is not None and self._hover_texts is not None:
            for i, line in enumerate(self._lines):
                if line.mouseShape().contains(ev.pos()):
                    vb.setToolTip(self._hover_texts[i])
                    self._toolTipCleared = False
                    break
            else:
                if not self._toolTipCleared:
                    vb.setToolTip("")
                    self._toolTipCleared = True

    def _plt_set_hover_text(self, texts: list[str]):
        self._hover_texts = texts

    def _plt_connect_pick_event(self, callback):
        self.clicked.connect(callback)



from __future__ import annotations

import numpy as np
import pyqtgraph as pg
from cmap import Colormap
from qtpy.QtGui import QTransform

from whitecanvas.backend.pyqtgraph._base import PyQtLayer
from whitecanvas.protocols import ImageProtocol, check_protocol


@check_protocol(ImageProtocol)
class Image(pg.ImageItem, PyQtLayer):
    def __init__(self, data: np.ndarray):
        super().__init__(np.swapaxes(data, 0, 1))
        self._cmap = Colormap("gray")
        self.setTransform(QTransform())

    def _plt_get_data(self) -> np.ndarray:
        return np.swapaxes(self.image, 0, 1)

    def _plt_set_data(self, data: np.ndarray):
        self.setImage(np.swapaxes(data, 0, 1))

    def _plt_get_colormap(self) -> Colormap:
        return self._cmap

    def _plt_set_colormap(self, cmap: Colormap):
        self._cmap = cmap
        self.setColorMap(cmap.to_pyqtgraph())

    def _plt_get_clim(self) -> tuple[float, float]:
        low, high = self.getLevels()
        return low, high

    def _plt_set_clim(self, clim: tuple[float, float]):
        self.setLevels(clim)

    def _get_qtransform(self) -> QTransform:
        return self.transform()

    def _plt_get_translation(self) -> tuple[float, float]:
        tr = self._get_qtransform()
        return tr.dx() + 0.5 * tr.m11(), tr.dy() + 0.5 * tr.m22()

    def _plt_set_translation(self, translation: tuple[float, float]):
        tr = self._get_qtransform()
        tr.setMatrix(
            tr.m11(),
            tr.m12(),
            tr.m13(),
            tr.m21(),
            tr.m22(),
            tr.m23(),
            translation[0] - 0.5 * tr.m11(),
            translation[1] - 0.5 * tr.m22(),
            tr.m33(),
        )
        self.setTransform(tr)

    def _plt_get_scale(self) -> tuple[float, float]:
        tr = self._get_qtransform()
        return tr.m11(), tr.m22()

    def _plt_set_scale(self, scale: tuple[float, float]):
        tr = self._get_qtransform()
        shift = self._plt_get_translation()
        tr.setMatrix(
            scale[0],
            tr.m12(),
            tr.m13(),
            tr.m21(),
            scale[1],
            tr.m23(),
            shift[0] - 0.5 * scale[0],
            shift[1] - 0.5 * scale[1],
            tr.m33(),
        )
        self.setTransform(tr)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    Iterator,
    Literal,
    Sequence,
    TypeVar,
)

from whitecanvas import layers as _l
from whitecanvas import theme
from whitecanvas.backend import Backend
from whitecanvas.canvas._grid import CanvasGrid
from whitecanvas.canvas._linker import link_axes
from whitecanvas.layers import group as _lg
from whitecanvas.layers import tabular as _lt
from whitecanvas.types import (
    ArrayLike1D,
    ColormapType,
    ColorType,
    Hatch,
    HistBinType,
    HistogramKind,
    HistogramShape,
    KdeBandWidthType,
    Location,
    LocationStr,
    Orientation,
    OrientationLike,
    Symbol,
)

if TYPE_CHECKING:
    from typing_extensions import Self

    from whitecanvas.canvas import Canvas
    from whitecanvas.canvas import _namespaces as _ns
    from whitecanvas.canvas.dataframe import JointCatPlotter
    from whitecanvas.layers import _mixin
    from whitecanvas.layers.tabular._dataframe import DataFrameWrapper

    NStr = str | Sequence[str]

_DF = TypeVar("_DF")


_0_or_1 = Literal[0, 1]


class JointGrid(CanvasGrid):
    """
    Grid with a main (joint) canvas and two marginal canvases.

    The marginal canvases shares the x-axis and y-axis with the main canvas.
    """

    def __init__(
        self,
        loc: tuple[_0_or_1, _0_or_1] = (1, 0),
        palette: str | ColormapType | None = None,
        ratio: int = 4,
        backend: Backend | str | None = None,
    ):
        widths = [1, 1]
        heights = [1, 1]
        rloc, cloc = loc
        if rloc not in (0, 1) or cloc not in (0, 1):
            raise ValueError(f"Invalid location {loc!r}.")
        widths[rloc] = heights[cloc] = ratio
        super().__init__(widths, heights, backend=Backend(backend))
        self._main_canvas = self.add_canvas(rloc, cloc, palette=palette)
        self._x_canvas = self.add_canvas(1 - rloc, cloc)
        self._y_canvas = self.add_canvas(rloc, 1 - cloc)

        # flip the axes if needed
        if rloc == 0:
            self._x_canvas.y.flipped = True
            self._x_namespace_canvas = self._x_canvas
            self._main_canvas.x.ticks.visible = False
            self._title_namespace_canvas = self._main_canvas
        else:
            self._x_namespace_canvas = self._main_canvas
            self._x_canvas.x.ticks.visible = False
            self._title_namespace_canvas = self._x_canvas
        if cloc == 0:
            self._ynamespace_canvas = self._main_canvas
            self._y_canvas.y.ticks.visible = False
        else:
            self._y_canvas.x.flipped = True
            self._ynamespace_canvas = self._y_canvas
            self._main_canvas.y.ticks.visible = False

        self._backend_object._plt_set_spacings(10, 10)

        # link axes
        self._x_linker = link_axes([self._main_canvas.x, self._x_canvas.x])
        self._y_linker = link_axes([self._main_canvas.y, self._y_canvas.y])

        # joint plotter
        self._x_plotters = []
        self._y_plotters = []

    def _iter_x_plotters(self) -> Iterator[MarginalPlotter]:
        if len(self._x_plotters) == 0:
            yield MarginalHistPlotter(Orientation.VERTICAL)
        else:
            yield from self._x_plotters

    def _iter_y_plotters(self) -> Iterator[MarginalPlotter]:
        if len(self._y_plotters) == 0:
            yield MarginalHistPlotter(Orientation.HORIZONTAL)
        else:
            yield from self._y_plotters

    @property
    def x_canvas(self) -> Canvas:
        """The canvas at the x-axis."""
        return self._x_canvas

    @property
    def y_canvas(self) -> Canvas:
        """The canvas at the y-axis."""
        return self._y_canvas

    @property
    def main_canvas(self) -> Canvas:
        """The main (joint) canvas."""
        return self._main_canvas

    @property
    def x(self) -> _ns.XAxisNamespace:
        """The x-axis namespace of the joint grid."""
        return self._x_namespace_canvas.x

    @property
    def y(self) -> _ns.YAxisNamespace:
        """The y-axis namespace of the joint grid."""
        return self._ynamespace_canvas.y

    @property
    def title(self) -> _ns.TitleNamespace:
        """Title namespace of the joint grid."""
        return self._title_namespace_canvas.title

    def cat(
        self,
        data: _DF,
        x: str | None = None,
        y: str | None = None,
        *,
        update_labels: bool = True,
    ) -> JointCatPlotter[Self, _DF]:
        """Create a joint categorical canvas."""
        from whitecanvas.canvas.dataframe import JointCatPlotter

        return JointCatPlotter(self, data, x, y, update_labels=update_labels)

    def _link_marginal_to_main(self, layer: _l.Layer, main: _l.Layer) -> None:
        # TODO: this is not the only thing to be done
        main.events.visible.connect_setattr(layer, "visible", maxargs=1)

    def add_legend(
        self,
        layers: Sequence[str | _l.Layer] | None = None,
        location: Location | LocationStr = "top_right",
        *,
        title: str | None = None,
    ):
        """Add legend to the main canvas."""
        self.main_canvas.add_legend(layers, location=location, title=title)
        return None

    def add_markers(
        self,
        xdata: ArrayLike1D,
        ydata: ArrayLike1D,
        *,
        name: str | None = None,
        symbol: Symbol | str | None = None,
        size: float | None = None,
        color: ColorType | None = None,
        alpha: float = 1.0,
        hatch: str | Hatch | None = None,
    ) -> _l.Markers[_mixin.ConstFace, _mixin.ConstEdge, float]:
        color = self._main_canvas._generate_colors(color)
        out = self._main_canvas.add_markers(
            xdata, ydata, name=name, symbol=symbol, size=size, color=color,
            alpha=alpha, hatch=hatch,
        )  # fmt: skip
        for _x_plt in self._iter_x_plotters():
            xlayer = _x_plt.add_layer_for_markers(
                xdata, color, hatch, backend=self._backend
            )
            self.x_canvas.add_layer(xlayer)
            self._link_marginal_to_main(xlayer, out)
        for _y_plt in self._iter_y_plotters():
            ylayer = _y_plt.add_layer_for_markers(
                ydata, color, hatch, backend=self._backend
            )
            self.y_canvas.add_layer(ylayer)
            self._link_marginal_to_main(ylayer, out)
        self._autoscale_layers()
        return out

    def with_hist_x(
        self,
        *,
        bins: HistBinType = "auto",
        limits: tuple[float, float] | None = None,
        kind: str | HistogramKind = HistogramKind.density,
        shape: str | HistogramShape = HistogramShape.bars,
    ) -> Self:
        """
        Configure the x-marginal canvas to have a histogram.

        Parameters
        ----------
        bins : int or 1D array-like, default "auto"
            Bins of the histogram. This parameter will directly be passed
            to `np.histogram`.
        limits : (float, float), optional
            Limits in which histogram will be built. This parameter will equivalent to
            the `range` paraneter of `np.histogram`.
        shape : {"step", "polygon", "bars"}, default "bars"
            Shape of the histogram. This parameter defines how to convert the data into
            the line nodes.
        kind : {"count", "density", "probability", "frequency", "percent"}, optional
            Kind of the histogram.
        """
        self._x_plotters.append(
            MarginalHistPlotter(
                Orientation.VERTICAL, bins=bins, limits=limits, kind=kind, shape=shape
            )
        )
        return self

    def with_hist_y(
        self,
        *,
        bins: HistBinType = "auto",
        limits: tuple[float, float] | None = None,
        kind: str | HistogramKind = HistogramKind.density,
        shape: str | HistogramShape = HistogramShape.bars,
    ) -> Self:
        """
        Configure the y-marginal canvas to have a histogram.

        Parameters
        ----------
        bins : int or 1D array-like, default "auto"
            Bins of the histogram. This parameter will directly be passed
            to `np.histogram`.
        limits : (float, float), optional
            Limits in which histogram will be built. This parameter will equivalent to
            the `range` paraneter of `np.histogram`.
        shape : {"step", "polygon", "bars"}, default "bars"
            Shape of the histogram. This parameter defines how to convert the data into
            the line nodes.
        kind : {"count", "density", "probability", "frequency", "percent"}, optional
            Kind of the histogram.
        """
        self._y_plotters.append(
            MarginalHistPlotter(
                Orientation.HORIZONTAL, bins=bins, limits=limits, kind=kind, shape=shape
            )
        )
        return self

    def with_hist(
        self,
        *,
        bins: HistBinType | tuple[HistBinType, HistBinType] = "auto",
        limits: tuple[float, float] | None = None,
        kind: str | HistogramKind = HistogramKind.density,
        shape: str | HistogramShape = HistogramShape.bars,
    ) -> Self:
        """
        Configure both of the marginal canvases to have histograms.

        Parameters
        ----------
        bins : int or 1D array-like, default "auto"
            Bins of the histogram. This parameter will directly be passed
            to `np.histogram`.
        limits : (float, float), optional
            Limits in which histogram will be built. This parameter will equivalent to
            the `range` paraneter of `np.histogram`.
        shape : {"step", "polygon", "bars"}, default "bars"
            Shape of the histogram. This parameter defines how to convert the data into
            the line nodes.
        kind : {"count", "density", "probability", "frequency", "percent"}, optional
            Kind of the histogram.
        """
        if isinstance(bins, tuple):
            bins_x, bins_y = bins
        else:
            bins_x = bins_y = bins
        self.with_hist_x(bins=bins_x, limits=limits, kind=kind, shape=shape)
        self.with_hist_y(bins=bins_y, limits=limits, kind=kind, shape=shape)
        return self

    def with_kde_x(
        self,
        *,
        width: float | None = None,
        band_width: KdeBandWidthType = "scott",
        fill_alpha: float = 0.2,
    ) -> Self:
        """
        Configure the x-marginal canvas to have a kernel density estimate (KDE) plot.

        Parameters
        ----------
        width : float, optional
            Width of the line. Use theme default if not specified.
        band_width : "scott", "silverman" or float, default "scott"
            Bandwidth of the kernel.
        fill_alpha : float, default 0.2
            Alpha value of the fill color.
        """
        width = theme._default("line.width", width)
        self._x_plotters.append(
            MarginalKdePlotter(
                Orientation.VERTICAL,
                width=width,
                band_width=band_width,
                fill_alpha=fill_alpha,
            )
        )
        return self

    def with_kde_y(
        self,
        *,
        width: float | None = None,
        band_width: KdeBandWidthType = "scott",
        fill_alpha: float = 0.2,
    ) -> Self:
        """
        Configure the y-marginal canvas to have a kernel density estimate (KDE) plot.

        Parameters
        ----------
        width : float, optional
            Width of the line. Use theme default if not specified.
        band_width : "scott", "silverman" or float, default "scott"
            Bandwidth of the kernel.
        fill_alpha : float, default 0.2
            Alpha value of the fill color.
        """
        width = theme._default("line.width", width)
        self._y_plotters.append(
            MarginalKdePlotter(
                Orientation.HORIZONTAL,
                width=width,
                band_width=band_width,
                fill_alpha=fill_alpha,
            )
        )
        return self

    def with_kde(
        self,
        *,
        width: float | None = None,
        band_width: KdeBandWidthType = "scott",
        fill_alpha: float = 0.2,
    ) -> Self:
        """
        Configure both of the marginal canvases to have KDE plots.

        Parameters
        ----------
        width : float, optional
            Width of the line. Use theme default if not specified.
        band_width : "scott", "silverman" or float, default "scott"
            Bandwidth of the kernel.
        fill_alpha : float, default 0.2
            Alpha value of the fill color.
        """
        self.with_kde_x(width=width, band_width=band_width, fill_alpha=fill_alpha)
        self.with_kde_y(width=width, band_width=band_width, fill_alpha=fill_alpha)
        return self

    def with_rug_x(self, *, width: float | None = None) -> Self:
        """
        Configure the x-marginal canvas to have a rug plot.

        Parameters
        ----------
        width : float, optional
            Width of the line. Use theme default if not specified.
        """
        width = theme._default("line.width", width)
        self._x_plotters.append(MarginalRugPlotter(Orientation.VERTICAL, width=width))
        return self

    def with_rug_y(self, *, width: float | None = None) -> Self:
        """
        Configure the y-marginal canvas to have a rug plot.

        Parameters
        ----------
        width : float, optional
            Width of the line. Use theme default if not specified.
        """
        width = theme._default("line.width", width)
        self._y_plotters.append(MarginalRugPlotter(Orientation.HORIZONTAL, width=width))
        return self

    def with_rug(self, *, width: float | None = None) -> Self:
        """
        Configure both of the marginal canvases to have rug plots.

        Parameters
        ----------
        width : float, optional
            Width of the line. Use theme default if not specified.
        """
        self.with_rug_x(width=width)
        self.with_rug_y(width=width)
        return self

    def _autoscale_layers(self):
        for layer in self.x_canvas.layers:
            if isinstance(layer, (_l.Rug, _lt.DFRug)):
                ylow, yhigh = self.x_canvas.y.lim
                layer.update_length((yhigh - ylow) * 0.1)
        for layer in self.y_canvas.layers:
            if isinstance(layer, (_l.Rug, _lt.DFRug)):
                xlow, xhigh = self.y_canvas.x.lim
                layer.update_length((xhigh - xlow) * 0.1)


class MarginalPlotter(ABC):
    def __init__(self, orient: OrientationLike):
        self._orient = Orientation.parse(orient)

    @abstractmethod
    def add_layer_for_markers(
        self,
        data: ArrayLike1D,
        color: ColorType,
        hatch: Hatch = Hatch.SOLID,
        backend: str | Backend | None = None,
    ) -> _l.Layer: ...

    @abstractmethod
    def add_layer_for_cat_markers(
        self,
        df: DataFrameWrapper[_DF],
        value: str,
        color: NStr | None = None,
        hatch: NStr | None = None,
        backend: str | Backend | None = None,
    ) -> _l.Layer: ...

    @abstractmethod
    def add_layer_for_cat_hist2d(
        self,
        df: DataFrameWrapper[_DF],
        value: str,
        color: str | None = None,
        bins: HistBinType | tuple[HistBinType, HistBinType] = "auto",
        limits: tuple[float, float] | None = None,
        backend: str | Backend | None = None,
    ) -> _l.Layer: ...


class MarginalHistPlotter(MarginalPlotter):
    def __init__(
        self,
        orient: OrientationLike,
        bins: HistBinType = "auto",
        limits: tuple[float, float] | None = None,
        kind: str | HistogramKind = "density",
        shape: str | HistogramShape = "bars",
    ):
        super().__init__(orient)
        self._bins = bins
        self._limits = limits
        self._kind = HistogramKind(kind)
        self._shape = HistogramShape(shape)

    def add_layer_for_markers(
        self,
        data: ArrayLike1D,
        color: ColorType,
        hatch: Hatch = Hatch.SOLID,
        backend: str | Backend | None = None,
    ) -> _lg.Histogram:
        return _lg.Histogram.from_array(
            data,
            shape=self._shape,
            kind=self._kind,
            color=color,
            orient=self._orient,
            bins=self._bins,
            limits=self._limits,
            backend=backend,
        )

    def add_layer_for_cat_markers(
        self,
        df: DataFrameWrapper[_DF],
        value: str,
        color: NStr | None = None,
        hatch: NStr | None = None,
        backend: str | Backend | None = None,
    ) -> _lt.DFHistograms[_DF]:
        return _lt.DFHistograms.from_table(
            df, value, orient=self._orient, color=color, hatch=hatch,
            bins=self._bins, limits=self._limits, kind=self._kind, shape=self._shape,
            backend=backend,
        )  # fmt: skip

    def add_layer_for_cat_hist2d(
        self,
        df: DataFrameWrapper[_DF],
        value: str,
        color: str | None = None,
        bins: HistBinType | tuple[HistBinType, HistBinType] = "auto",
        limits: tuple[float, float] | None = None,
        backend: str | Backend | None = None,
    ) -> _lt.DFHistograms[_DF]:
        if self._bins != "auto":
            bins = self._bins
        if self._limits is not None:
            limits = self._limits
        return _lt.DFHistograms.from_table(
            df, value, orient=self._orient, color=color, bins=bins, limits=limits,
            kind=self._kind, shape=self._shape, backend=backend,
        )  # fmt: skip


class MarginalKdePlotter(MarginalPlotter):
    def __init__(
        self,
        orient: OrientationLike,
        width: float = 1.0,
        band_width: KdeBandWidthType = "scott",
        fill_alpha: float = 0.2,
    ):
        super().__init__(orient)
        self._width = width
        self._band_width = band_width
        self._fill_alpha = fill_alpha

    def add_layer_for_markers(
        self,
        data: ArrayLike1D,
        color: ColorType,
        hatch: Hatch = Hatch.SOLID,
        backend: str | Backend | None = None,
    ) -> _lg.Kde:
        out = _lg.Kde.from_array(
            data, color=color, orient=self._orient, band_width=self._band_width,
            width=self._width, backend=backend,
        )  # fmt: skip
        out.fill_alpha = self._fill_alpha
        return out

    def add_layer_for_cat_markers(
        self,
        df: DataFrameWrapper[_DF],
        value: str,
        color: NStr | None = None,
        hatch: NStr | None = None,
        backend: str | Backend | None = None,
    ) -> _lt.DFKde[_DF]:
        out = _lt.DFKde.from_table(
            df, value, orient=self._orient, color=color, hatch=hatch,
            width=self._width, backend=backend,
        )  # fmt: skip
        for layer in out.base:
            layer.fill_alpha = self._fill_alpha
        return out

    def add_layer_for_cat_hist2d(
        self,
        df: DataFrameWrapper[_DF],
        value: str,
        color: str | None = None,
        bins: HistBinType | tuple[HistBinType, HistBinType] = "auto",
        limits: tuple[float, float] | None = None,
        backend: str | Backend | None = None,
    ) -> _lt.DFKde[_DF]:
        out = _lt.DFKde.from_table(
            df, value, orient=self._orient, color=color, band_width=self._band_width,
            width=self._width, backend=backend,
        )  # fmt: skip
        for layer in out.base:
            layer.fill_alpha = self._fill_alpha
        return out


class MarginalRugPlotter(MarginalPlotter):
    def __init__(
        self,
        orient: OrientationLike,
        width: float = 1.0,
        length: float = 0.1,
    ):
        super().__init__(orient)
        self._width = width
        self._length = length

    def add_layer_for_markers(
        self,
        data: ArrayLike1D,
        color: ColorType,
        hatch: Hatch = Hatch.SOLID,
        backend: str | Backend | None = None,
    ) -> _l.Rug:
        return _l.Rug(
            data, high=self._length, color=color, orient=self._orient,
            width=self._width, backend=backend
        )  # fmt: skip

    def add_layer_for_cat_markers(
        self,
        df: DataFrameWrapper[_DF],
        value: str,
        color: NStr | None = None,
        hatch: NStr | None = None,
        backend: str | Backend | None = None,
    ) -> _lt.DFRug[_DF]:
        return _lt.DFRug.from_table(
            df,
            value,
            high=self._length,
            orient=self._orient,
            color=color,
            width=self._width,
            backend=backend,
        )

    def add_layer_for_cat_hist2d(
        self,
        df: DataFrameWrapper[_DF],
        value: str,
        color: str | None = None,
        bins: HistBinType | tuple[HistBinType, HistBinType] = "auto",
        limits: tuple[float, float] | None = None,
        backend: str | Backend | None = None,
    ) -> _lt.DFRug[_DF]:
        return _lt.DFRug.from_table(
            df,
            value,
            high=self._length,
            orient=self._orient,
            color=color,
            width=self._width,
            backend=backend,
        )


# class MarginalBoxPlotter
# class MarginalViolinPlotter



from __future__ import annotations

from typing import Any

from whitecanvas.utils.normalize import as_array_1d
from whitecanvas.utils.type_check import is_real_number


class TextOffset:
    def _asarray(self) -> tuple[Any, Any]:
        raise NotImplementedError

    def _add(self, dx, dy) -> TextOffset:
        raise NotImplementedError


class NoOffset(TextOffset):
    def _asarray(self):
        return (0, 0)

    def _add(self, dx, dy) -> TextOffset:
        if is_real_number(dx) and is_real_number(dy):
            if dx == 0 and dy == 0:
                return self
            return ConstantOffset(dx, dy)
        else:
            return CustomOffset(as_array_1d(dx), as_array_1d(dy))

    def __repr__(self) -> str:
        return "<NoOffset>"


class ConstantOffset(TextOffset):
    def __init__(self, x: float, y: float):
        self._x, self._y = x, y

    def _asarray(self) -> tuple[float, float]:
        return (self._x, self._y)

    def _add(self, dx, dy) -> TextOffset:
        if is_real_number(dx) and is_real_number(dy):
            return ConstantOffset(self._x + dx, self._y + dy)
        else:
            return CustomOffset(as_array_1d(dx) + self._x, as_array_1d(dy) + self._y)

    def __repr__(self) -> str:
        return f"<ConstantOffset({self._x}, {self._y})>"


class CustomOffset(TextOffset):
    def __init__(self, x: Any, y: Any):
        self._x, self._y = x, y

    def _asarray(self) -> tuple[Any, Any]:
        return (self._x, self._y)

    def _add(self, dx, dy) -> TextOffset:
        return CustomOffset(as_array_1d(dx + self._x), as_array_1d(dy + self._y))

    def __repr__(self) -> str:
        return f"<CustomOffset({self._x}, {self._y})>"



from __future__ import annotations

import weakref
from typing import (
    TYPE_CHECKING,
    Generic,
    Iterable,
    Iterator,
    Literal,
    MutableSequence,
    Sequence,
    TypeVar,
    overload,
)

from psygnal import Signal

from whitecanvas.layers import _legend
from whitecanvas.layers._base import Layer, LayerEvents, LayerGroup, PrimitiveLayer

if TYPE_CHECKING:
    from whitecanvas.canvas import Canvas

_L = TypeVar("_L", bound=PrimitiveLayer)


class LayerContainerEvents(LayerEvents):
    reordered = Signal(list)


class RichContainerEvents(LayerContainerEvents):
    face = Signal(object)
    edge = Signal(object)


class LayerContainer(LayerGroup):
    """Layer group that stores children in a list."""

    events: LayerContainerEvents
    _events_class = LayerContainerEvents

    def __init__(self, children: Iterable[Layer], name: str | None = None):
        super().__init__(name=name)
        self._children = [_process_grouping(c, self) for c in children]
        self._ordering_indices = self._default_ordering(len(self._children))
        self._emit_layer_grouped()

    def _default_ordering(self, n: int) -> list[int]:
        """Return the default ordering of the children."""
        return list(range(n))

    def iter_children(self) -> Iterator[Layer]:
        """Recursively iterate over all children."""
        for _, child in sorted(
            zip(self._ordering_indices, self._children), key=lambda x: x[0]
        ):
            yield child

    @property
    def zorders(self) -> list[int]:
        """The z-orders of the children."""
        return list(self._ordering_indices)

    @zorders.setter
    def zorders(self, zorders: list[int]):
        zorders = list(zorders)
        if set(zorders) != set(range(len(self._children))):
            raise ValueError(
                "zorders must be a permutation of numbers from 0 to N-1, "
                f"like [0, 2, 1, 3], but got {zorders}"
            )
        self._ordering_indices = zorders
        self.events.reordered.emit(zorders)

    def _connect_canvas(self, canvas: Canvas):
        # TODO: connect plt.draw
        return super()._connect_canvas(canvas)

    def _disconnect_canvas(self, canvas: Canvas):
        return super()._disconnect_canvas(canvas)


class LayerTuple(LayerContainer, Sequence[Layer]):
    def __getitem__(self, key: int | str) -> Layer:
        if isinstance(key, str):
            for child in self.iter_children():
                if child.name == key:
                    return child
            raise ValueError(f"Layer {key!r} not found")
        return self._children[key]

    def __len__(self) -> int:
        return len(self._children)

    def __iter__(self) -> Iterator[Layer]:
        return self.iter_children()

    def __repr__(self) -> str:
        cname = type(self).__name__
        return f"{cname}{tuple(self.iter_children())!r}"

    def _as_legend_item(self) -> _legend.LegendItem:
        items = [(layer.name, layer._as_legend_item()) for layer in self]
        return _legend.LegendItemCollection(items)


class LayerCollectionBase(LayerContainer, MutableSequence[_L]):
    _children: list[_L]

    def __getitem__(self, n: int) -> _L:
        """The n-th layer."""
        if not hasattr(n, "__index__"):
            raise TypeError(f"Index must be an integer, not {type(n)}")
        return self._children[n]

    def __setitem__(self, n: int, layer: _L):
        raise NotImplementedError("Cannot set item in a LayerCollection")

    def __delitem__(self, n: int):
        """Delete the n-th layer."""
        if not hasattr(n, "__index__"):
            raise TypeError(f"Index must be an integer, not {type(n)}")
        line = self._children.pop(n)
        if _canvas := self._canvas_ref():
            _canvas._canvas()._plt_remove_layer(line._backend)
            line._disconnect_canvas(_canvas)
        return None

    def __iter__(self) -> Iterator[_L]:
        return iter(self._children)

    def __len__(self):
        return len(self._children)

    def insert(self, n: int, layer: _L):
        if _canvas := self._canvas_ref():
            _canvas._canvas()._plt_add_layer(layer._backend)
            layer._connect_canvas(_canvas)
        _process_grouping(layer, self)
        self._children.insert(n, layer)
        self._ordering_indices.insert(n, len(self._ordering_indices))
        return None

    def _as_legend_item(self):
        """Use the first layer as the main legend item."""
        if len(self) == 0:
            return _legend.EmptyLegendItem()
        return self[0]._as_legend_item()

    if TYPE_CHECKING:

        def iter_children(self) -> Iterator[_L]:
            ...


_L0 = TypeVar("_L0", bound=Layer)
_L1 = TypeVar("_L1", bound=Layer)


class MainAndOtherLayers(LayerTuple, Generic[_L0, _L1]):
    @overload
    def __getitem__(self, n: Literal[0]) -> _L0:
        ...

    @overload
    def __getitem__(self, n: Literal[0]) -> _L1:
        ...

    @overload
    def __getitem__(self, n: int) -> Layer:
        ...

    def __getitem__(self, n):
        """The n-th layer."""
        return super().__getitem__(n)

    def _insert(self, layer: Layer):
        if layer._canvas_ref() is not None:
            raise ValueError(f"{layer!r} is already added to a canvas")
        if _canvas := self._canvas_ref():
            if isinstance(layer, PrimitiveLayer):
                _canvas._canvas()._plt_add_layer(layer._backend)
            elif isinstance(layer, LayerGroup):
                for l in layer.iter_primitive():
                    _canvas._canvas()._plt_add_layer(l._backend)
            layer._connect_canvas(_canvas)
        _process_grouping(layer, self)
        self._children.insert(1, layer)
        self._ordering_indices.insert(1, len(self._ordering_indices))
        return None

    def _as_legend_item(self):
        """Use the first layer as the main legend item."""
        if len(self) == 0:
            # this should never happen, but just in case
            return _legend.EmptyLegendItem()
        return self[0]._as_legend_item()


def _process_grouping(l, parent: Layer) -> Layer:
    if not isinstance(l, Layer):
        raise TypeError(f"{l!r} is not a Layer")
    if l._group_layer_ref is not None:
        raise ValueError(f"{l!r} is already grouped")
    l._group_layer_ref = weakref.ref(parent)
    return l



from __future__ import annotations

import numpy as np

from whitecanvas.types import Hatch, LineStyle


def edge_style():
    def _getter(self):
        return getattr(self, "__edge_style_value", LineStyle.SOLID)

    def _setter(self, value: LineStyle):
        setattr(self, "__edge_style_value", value)

    return _getter, _setter


def edge_styles():
    def _getter(self):
        return getattr(
            self, "__edge_style_value", [LineStyle.SOLID] * self._plt_get_ndata()
        )

    def _setter(self, value: LineStyle | list[LineStyle]):
        if isinstance(value, LineStyle):
            value = [value] * self._plt_get_ndata()
        setattr(self, "__edge_style_value", value)

    return _getter, _setter


def edge_width():
    def _getter(self):
        return getattr(self, "__edge_width_value", 1.0)

    def _setter(self, value: float):
        setattr(self, "__edge_width_value", value)

    return _getter, _setter


def edge_color():
    def _getter(self):
        return getattr(
            self, "__edge_color_value", np.array([0, 0, 0, 255], dtype=np.float32)
        )

    def _setter(self, value: float):
        setattr(self, "__edge_color_value", value)

    return _getter, _setter


def face_hatch():
    def _getter(self):
        return getattr(self, "__face_hatch_value", Hatch.SOLID)

    def _setter(self, value: Hatch):
        setattr(self, "__face_hatch_value", value)

    return _getter, _setter


def face_hatches():
    def _getter(self):
        return getattr(
            self, "__face_hatch_value", [Hatch.SOLID] * self._plt_get_ndata()
        )

    def _setter(self, value: Hatch | list[Hatch]):
        if isinstance(value, Hatch):
            value = [value] * self._plt_get_ndata()
        setattr(self, "__face_hatch_value", value)

    return _getter, _setter



from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING

import numpy as np
import pyqtgraph as pg
from numpy.typing import NDArray
from qtpy import QtGui

from whitecanvas.backend.pyqtgraph._base import PyQtLayer
from whitecanvas.backend.pyqtgraph._qt_utils import (
    array_to_qcolor,
    from_qt_brush_style,
    from_qt_line_style,
    to_qt_brush_style,
    to_qt_line_style,
)
from whitecanvas.protocols import TextProtocol, check_protocol
from whitecanvas.types import Alignment, Hatch, LineStyle
from whitecanvas.utils.normalize import as_color_array
from whitecanvas.utils.type_check import is_real_number


@check_protocol(TextProtocol)
class Texts(pg.ItemGroup, PyQtLayer):
    def __init__(
        self, x: NDArray[np.floating], y: NDArray[np.floating], text: list[str]
    ):
        super().__init__()
        for x0, y0, text0 in zip(x, y, text):
            self.addItem(SingleText(x0, y0, text0))
        self._font_family = "Arial"
        self._align = Alignment.BOTTOM_LEFT

    if TYPE_CHECKING:

        def childItems(self) -> list[SingleText]:
            ...

    ##### TextProtocol #####

    def _plt_get_text(self) -> list[str]:
        return [t.toPlainText() for t in self.childItems()]

    def _plt_set_text(self, text: list[str]):
        for t, text0 in zip(self.childItems(), text):
            t.setPlainText(text0)

    def _plt_get_ndata(self) -> int:
        return len(self.childItems())

    def _plt_get_text_color(self):
        return np.array([t.color.getRgbF() for t in self.childItems()])

    def _plt_set_text_color(self, color):
        color = as_color_array(color, self._plt_get_ndata())
        for t, color0 in zip(self.childItems(), color):
            t.setColor(array_to_qcolor(color0))

    def _plt_get_text_size(self) -> float:
        return [t._plt_get_text_size() for t in self.childItems()]

    def _plt_set_text_size(self, size: float | NDArray[np.floating]):
        if is_real_number(size):
            size = np.full(self._plt_get_ndata(), size)
        for t, size0 in zip(self.childItems(), size):
            t._plt_set_text_size(size0)

    def _plt_get_text_position(
        self,
    ) -> tuple[NDArray[np.floating], NDArray[np.floating]]:
        pos = np.array([t.pos() for t in self.childItems()])
        if pos.size == 0:
            return np.zeros(0, dtype=np.float32), np.zeros(0, dtype=np.float32)
        return pos[:, 0], pos[:, 1]

    def _plt_set_text_position(
        self, position: tuple[NDArray[np.floating], NDArray[np.floating]]
    ):
        x, y = position
        if x.size > self._plt_get_ndata():
            for _ in range(x.size - self._plt_get_ndata()):
                self.addItem(SingleText(0, 0, ""))
        elif x.size < self._plt_get_ndata():
            for _ in range(self._plt_get_ndata() - x.size):
                self.childItems()[-1].setParentItem(None)
        for t, x0, y0 in zip(self.childItems(), x, y):
            t.setPos(x0, y0)

    def _plt_get_text_anchor(self) -> list[Alignment]:
        return self._align

    def _plt_set_text_anchor(self, anc: Alignment):
        for t in self.childItems():
            t._plt_set_text_anchor(anc)
        self._align = anc

    def _plt_get_text_rotation(self) -> float:
        return [t.angle for t in self.childItems()]

    def _plt_set_text_rotation(self, rotation: float):
        if is_real_number(rotation):
            rotation = np.full(self._plt_get_ndata(), rotation)
        for t, rotation0 in zip(self.childItems(), rotation):
            t.setAngle(rotation0)

    def _plt_get_text_fontfamily(self) -> str:
        return self._font_family

    def _plt_set_text_fontfamily(self, fontfamily: str):
        for t in self.childItems():
            font = t._get_qfont()
            font.setFamily(fontfamily)
            t.setFont(font)
        self._font_family = fontfamily

    ##### HasFaces #####

    def _plt_get_face_color(self):
        return np.array([t._get_brush().color().getRgbF() for t in self.childItems()])

    def _plt_set_face_color(self, color):
        color = as_color_array(color, self._plt_get_ndata())
        for t, color0 in zip(self.childItems(), color):
            brush = t._get_brush()
            brush.setColor(array_to_qcolor(color0))
            t.fill = brush

    def _plt_get_face_hatch(self) -> list[Hatch]:
        return [from_qt_brush_style(s._get_brush().style()) for s in self.childItems()]

    def _plt_set_face_hatch(self, pattern: Hatch | list[Hatch]):
        if isinstance(pattern, Hatch):
            pattern = [pattern] * self._plt_get_ndata()
        for t, pattern0 in zip(self.childItems(), pattern):
            brush = t._get_brush()
            brush.setStyle(to_qt_brush_style(pattern0))
            t.fill = brush

    ##### HasEdges #####

    def _plt_get_edge_color(self):
        return np.array(
            [t._get_pen().color().getRgbF() for t in self.childItems()],
            dtype=np.float32,
        )

    def _plt_set_edge_color(self, color):
        color = as_color_array(color, self._plt_get_ndata())
        for t, color0 in zip(self.childItems(), color):
            pen = t._get_pen()
            pen.setColor(array_to_qcolor(color0))
            t.border = pen

    def _plt_get_edge_width(self) -> float:
        return np.array(
            [t._get_pen().widthF() for t in self.childItems()], dtype=np.float32
        )

    def _plt_set_edge_width(self, width: float | NDArray[np.floating]):
        if is_real_number(width):
            width = np.full(self._plt_get_ndata(), width)
        for t, width0 in zip(self.childItems(), width):
            pen = t._get_pen()
            pen.setWidthF(width0)
            t.border = pen

    def _plt_get_edge_style(self) -> list[LineStyle]:
        return [from_qt_line_style(s._get_pen().style()) for s in self.childItems()]

    def _plt_set_edge_style(self, style: LineStyle | list[LineStyle]):
        if isinstance(style, LineStyle):
            style = [style] * self._plt_get_ndata()
        for t, style0 in zip(self.childItems(), style):
            pen = t._get_pen()
            pen.setStyle(to_qt_line_style(style0))
            t.border = pen


class SingleText(pg.TextItem, PyQtLayer):
    color: QtGui.QColor

    def __init__(self, x: float, y: float, text: str):
        super().__init__(text=text, anchor=(0, 0))
        self._plt_set_text_position([x, y])
        self._alignment = Alignment.BOTTOM_LEFT

    ##### TextProtocol #####

    def _get_qfont(self) -> QtGui.QFont:
        return self.textItem.font()

    def _plt_get_text_size(self) -> float:
        return self._get_qfont().pointSize()

    def _plt_set_text_size(self, size: float):
        font = self._get_qfont()
        font.setPointSizeF(size)
        self.setFont(font)

    def _plt_get_text_position(self) -> tuple[float, float]:
        return self.pos()

    def _plt_set_text_position(self, position: tuple[float, float]):
        self.setPos(*position)

    def _plt_get_text_anchor(self) -> Alignment:
        return self._alignment

    def _plt_set_text_anchor(self, anc: Alignment):
        self.setAnchor(_split_anchor(anc))
        self._alignment = anc

    def _get_brush(self) -> QtGui.QBrush:
        return self.fill

    def _get_pen(self) -> QtGui.QPen:
        return self.border


@lru_cache(maxsize=9)
def _split_anchor(anc: Alignment):
    va, ha = anc.split()
    if va is Alignment.TOP:
        yanc = 0
    elif va is Alignment.BOTTOM:
        yanc = 1
    else:
        yanc = 0.5
    if ha is Alignment.LEFT:
        xanc = 0
    elif ha is Alignment.RIGHT:
        xanc = 1
    else:
        xanc = 0.5
    return xanc, yanc



from __future__ import annotations

import sys
from enum import Enum
from numbers import Real
from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    import pandas as pd
    import polars as pl
    from typing_extensions import TypeGuard


def is_not_array(x) -> bool:
    return np.isscalar(x) or isinstance(x, Enum)


def is_real_number(x) -> TypeGuard[float]:
    return isinstance(x, (int, float, Real, np.number))


def is_pandas_dataframe(df) -> TypeGuard[pd.DataFrame]:
    typ = type(df)
    if (
        typ.__name__ != "DataFrame"
        or "pandas" not in sys.modules
        or typ.__module__.split(".")[0] != "pandas"
    ):
        return False
    import pandas as pd

    return isinstance(df, pd.DataFrame)


def is_polars_dataframe(df) -> TypeGuard[pl.DataFrame]:
    typ = type(df)
    if (
        typ.__name__ != "DataFrame"
        or "polars" not in sys.modules
        or typ.__module__.split(".")[0] != "polars"
    ):
        return False
    import polars as pl

    return isinstance(df, pl.DataFrame)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Generic, Sequence, TypeVar

import numpy as np

from whitecanvas.canvas.dataframe import _misc
from whitecanvas.canvas.dataframe._base import BaseCatPlotter, CatIterator
from whitecanvas.layers import tabular as _lt
from whitecanvas.layers.tabular import _jitter
from whitecanvas.types import ColormapType

if TYPE_CHECKING:
    from typing_extensions import Self

    from whitecanvas.canvas._base import CanvasBase
    from whitecanvas.layers.tabular._dataframe import DataFrameWrapper

    NStr = str | Sequence[str]

_C = TypeVar("_C", bound="CanvasBase")
_DF = TypeVar("_DF")


class _XYAggregator(Generic[_C, _DF]):
    """Aggregator for XY-categorical plots."""

    def __init__(self, method: str, plotter: XYCatPlotter[_C, _DF] = None):
        self._method = method
        self._plotter = plotter

    def __get__(self, ins: _C, owner) -> Self:
        return _XYAggregator(self._method, ins)

    def __repr__(self) -> str:
        return f"XYAggregator<{self._method}>"

    def __call__(self) -> XYCatAggPlotter[_C, _DF]:
        """
        Aggregate the values before plotting it.

        Examples
        --------
        >>> canvas.cat_xy(df, "x", "y").count().add_heatmap("value")
        >>> canvas.cat_xy(df, "x", "y").mean().add_heatmap("value")
        """
        plotter = self._plotter
        if plotter is None:
            raise TypeError("Cannot call this method from a class.")
        if plotter._x is None or plotter._y is None:
            raise ValueError("Value column is not specified.")
        return XYCatAggPlotter(
            plotter._canvas(),
            plotter._cat_iter_x,
            plotter._cat_iter_y,
            x=plotter._x,
            y=plotter._y,
            method=self._method,
        )


class XYCatPlotter(BaseCatPlotter[_C, _DF]):
    def __init__(
        self,
        canvas: _C,
        df: _DF,
        x: str | tuple[str, ...],
        y: str | tuple[str, ...],
        update_labels: bool = False,
        x_sorter: Callable[[tuple], int] | None = None,
        y_sorter: Callable[[tuple], int] | None = None,
    ):
        super().__init__(canvas, df)
        if isinstance(x, str):
            x = (x,)
        if isinstance(y, str):
            y = (y,)
        self._x: tuple[str, ...] = x
        self._y: tuple[str, ...] = y
        self._update_label = update_labels
        self._cat_iter_x = CatIterator(self._df, x, sort_func=x_sorter)
        self._cat_iter_y = CatIterator(self._df, y, sort_func=y_sorter)
        if update_labels:
            self._update_xy_label(x, y)
        self._update_axis_labels()

    def sort_in_order(
        self,
        *,
        x: Sequence[Any] | None = None,
        y: Sequence[Any] | None = None,
    ) -> Self:
        """
        Sort the data in the given order.

        Parameters
        ----------
        x : sequence of any
            The order to sort the x-axis data.
        y : sequence of any
            The order to sort the y-axis data.
        """
        if x is not None:
            x_sorter = _misc.make_sorter_manual(
                [u if isinstance(u, tuple) else (u,) for u in x]
            )
        else:
            x_sorter = None
        if y is not None:
            y_sorter = _misc.make_sorter_manual(
                [u if isinstance(u, tuple) else (u,) for u in y]
            )
        else:
            y_sorter = None

        return type(self)(
            self._canvas(),
            self._df,
            x=self._x,
            y=self._y,
            update_labels=self._update_label,
            x_sorter=x_sorter,
            y_sorter=y_sorter,
        )

    def sort(self, *, x_ascending: bool = True, y_ascending: bool = True) -> Self:
        """
        Sort the categories in ascending or descending order.

        Parameters
        ----------
        x_ascending : bool, default True
            Sort the x-axis categories in ascending order.
        y_ascending : bool, default True
            Sort the y-axis categories in ascending order.
        """
        return type(self)(
            self._canvas(),
            self._df,
            x=self._x,
            y=self._y,
            update_labels=self._update_label,
            x_sorter=_misc.sorter_ascending if x_ascending else _misc.sorter_descending,
            y_sorter=_misc.sorter_ascending if y_ascending else _misc.sorter_descending,
        )

    def _update_xy_label(
        self,
        x: str | tuple[str, ...],
        y: str | tuple[str, ...],
    ) -> None:
        """Update the x and y labels using the column names"""
        canvas = self._canvas()
        if not isinstance(x, str):
            x = "/".join(x)
        if not isinstance(y, str):
            y = "/".join(y)
        canvas.x.label.text = x
        canvas.y.label.text = y

    def _update_axis_labels(self) -> None:
        """Update the x and y labels using the column names"""
        canvas = self._canvas()
        canvas.x.ticks.set_labels(*self._cat_iter_x.axis_ticks())
        canvas.y.ticks.set_labels(*self._cat_iter_y.axis_ticks())

    mean = _XYAggregator("mean")
    median = _XYAggregator("median")
    sum = _XYAggregator("sum")
    min = _XYAggregator("min")
    max = _XYAggregator("max")
    count = _XYAggregator("size")
    first = _XYAggregator("first")


class XYCatAggPlotter(BaseCatPlotter[_C, _DF]):
    def __init__(
        self,
        canvas: _C,
        cat_iter_x: CatIterator[_DF],
        cat_iter_y: CatIterator[_DF],
        x: str | tuple[str, ...],
        y: str | tuple[str, ...],
        method: str,
    ):
        super().__init__(canvas, cat_iter_x.df)
        self._cat_iter_x = cat_iter_x
        self._cat_iter_y = cat_iter_y
        self._x = x
        self._y = y
        self._agg_method = method

    def add_heatmap(
        self,
        value: str,
        *,
        cmap: ColormapType = "inferno",
        clim: tuple[float, float] | None = None,
        name: str | None = None,
        fill: float = 0,
    ) -> _lt.DFHeatmap[_DF]:
        """
        Add a heatmap whose color represents the value of the aggregated data.

        Parameters
        ----------
        value : str
            Column name to use as the value.
        cmap : colormap-like, default "inferno"
            Colormap to use for the heatmap.
        clim : (float, float), optional
            Color limits for the colormap. If not specified, the limits are calculated
            from the data min/max.
        name : str, optional
            Name of the layer.
        fill : float, optional
            Value to fill for the cells that do not have any data. This value will not
            be considered when calculating the color limits.

        Returns
        -------
        DFHeatmap
            Dataframe bound heatmap layer.
        """
        canvas = self._canvas()
        by_both = (*self._x, *self._y)
        nx = len(self._x)
        df_agg = self._aggregate(self._df, by_both, value)
        map_x = self._cat_iter_x.prep_position_map(self._x)
        map_y = self._cat_iter_y.prep_position_map(self._y)
        dtype = df_agg[value].dtype
        if dtype.kind not in "fiub":
            raise ValueError(f"Column {value!r} is not numeric.")
        arr = np.full((len(map_y), len(map_x)), fill, dtype=dtype)
        for sl, sub in df_agg.group_by(by_both):
            xval, yval = sl[:nx], sl[nx:]
            vals = sub[value]
            if vals.size == 1:
                arr[map_y[yval], map_x[xval]] = vals[0]
            else:
                raise ValueError(f"More than one value found for {sl!r}.")
        if clim is None:
            # `fill` may be outside the range of the data, so calculate clim here.
            clim = df_agg[value].min(), df_agg[value].max()
        layer = _lt.DFHeatmap.from_array(
            df_agg, arr, name=name, cmap=cmap, clim=clim, backend=canvas._get_backend(),
        )  # fmt: skip
        return canvas.add_layer(layer)

    def add_markers(
        self,
        value: str,
        *,
        color: NStr | None = None,
        name: str | None = None,
        map_from: tuple[float, float] | None = None,
        map_to: tuple[float, float] = (3, 18),
    ) -> _lt.DFMarkerGroups[_DF]:
        """
        Add markers at the grid positions with the size representing the value.

        >>> canvas.cat_xy(df, "x", "y").count().add_markers("value")

        Parameters
        ----------
        value : str
            Column name to use as the value.
        color : str or sequence of str, optional
            Column name(s) to use as the color of the markers.
        name : str, optional
            Name of the layer.
        map_from : tuple of float, optional
            Limits of values that will be linearly mapped to the marker size. Data
            points outside this range will be clipped. If not specified, the min/max
            of the data will be used.
        map_to : tuple of float, optional
            Minimum and maximum size of the markers.

        Returns
        -------
        _lt.DFMarkerGroups[_DF]
            _description_
        """
        canvas = self._canvas()
        x, y = self._x, self._y
        xj = _jitter.CategoricalJitter(x, self._cat_iter_x.prep_position_map(x))
        yj = _jitter.CategoricalJitter(y, self._cat_iter_y.prep_position_map(y))
        return canvas.add_layer(
            _lt.DFMarkers(self._df, xj, yj, color=color, name=name).update_size(
                value, map_from=map_from, map_to=map_to
            )
        )

    def _aggregate(
        self,
        df: DataFrameWrapper[_DF],
        by: tuple[str, ...],
        on: str,
    ) -> DataFrameWrapper[_DF]:
        if self._agg_method == "size":
            return df.value_count(by)
        elif self._agg_method == "first":
            return df.value_first(by, on)
        else:
            if on is None:
                raise ValueError("Value column is not specified.")
            return df.agg_by(by, [on], self._agg_method)



from __future__ import annotations

import weakref
from typing import TYPE_CHECKING

import numpy as np

from whitecanvas.types import LineStyle
from whitecanvas.utils.normalize import rgba_str_color

if TYPE_CHECKING:
    from plotly.graph_objs.layout import Title as PlotlyTitle

    from whitecanvas.backend.plotly.canvas import Canvas


class _CanvasComponent:
    def __init__(self, canvas: Canvas):
        self._canvas = weakref.ref(canvas)


class _SupportsTitle:
    def __init__(self, canvas: Canvas):
        self._canvas = weakref.ref(canvas)
        self._visible = True
        self._text = ""

    def _get_title(self) -> PlotlyTitle:
        raise NotImplementedError

    def _plt_get_visible(self) -> bool:
        return self._visible

    def _plt_set_visible(self, visible: bool):
        if visible:
            self._get_title().text = self._text
        else:
            self._get_title().text = ""
        self._visible = visible

    def _plt_get_text(self) -> str:
        return self._get_title().text

    def _plt_set_text(self, text: str):
        self._get_title().text = text
        self._text = text

    def _plt_get_color(self):
        return self._get_title().font.color

    def _plt_set_color(self, color):
        self._get_title().font.color = rgba_str_color(color)

    def _plt_get_size(self) -> int:
        return self._get_title().font.size

    def _plt_set_size(self, size: int):
        self._get_title().font.size = size

    def _plt_get_fontfamily(self) -> str:
        return self._get_title().font.family

    def _plt_set_fontfamily(self, family: str):
        self._get_title().font.family = family


class Title(_SupportsTitle):
    def __init__(self, canvas: Canvas):
        super().__init__(canvas)
        canvas._fig.layout.title = {"text": "", "x": 0.5, "xanchor": "center"}

    def _get_title(self):
        return self._canvas()._fig.layout.title


class AxisLabel(_SupportsTitle):
    def __init__(self, canvas: Canvas, axis: str):
        super().__init__(canvas)
        self._axis = axis

    def _get_title(self):
        layout = self._canvas()._subplot_layout()
        return getattr(layout, self._axis).title


class AxisLabel3D(AxisLabel):
    def _get_title(self):
        layout = self._canvas()._subplot_layout()
        return getattr(layout.scene, self._axis).title


class Axis(_CanvasComponent):
    def __init__(self, canvas: Canvas, axis: str):
        super().__init__(canvas)
        self._axis = axis

    @property
    def name(self) -> str:
        return self._axis

    def _plt_get_axis(self):
        return getattr(self._canvas()._subplot_layout(), self._axis)

    def _plt_get_limits(self) -> tuple[float, float]:
        lim = self._plt_get_axis().range
        if lim is None:
            lim = (0, 1)  # TODO: how to get the limits?
        return lim

    def _plt_set_limits(self, limits: tuple[float, float]):
        self._plt_get_axis().range = limits

    def _plt_get_color(self):
        # color of the axis itself
        return np.array(self._plt_get_axis().linecolor)

    def _plt_set_color(self, color):
        self._plt_get_axis().linecolor = rgba_str_color(color)

    def _plt_flip(self) -> None:
        if self._plt_get_axis().autorange is None:
            self._plt_get_axis().autorange = "reversed"
        else:
            self._plt_get_axis().autorange = None

    def _plt_set_grid_state(self, visible: bool, color, width: float, style: LineStyle):
        axis = self._plt_get_axis()
        axis.showgrid = visible
        axis.gridcolor = rgba_str_color(color)
        axis.gridwidth = width


class Axis3D(Axis):
    def _plt_get_axis(self):
        return getattr(self._canvas()._subplot_layout().scene, self._axis)


class Ticks(_CanvasComponent):
    def __init__(self, canvas: Canvas, axis: str):
        super().__init__(canvas)
        self._axis = axis
        self._visible = True

    def _plt_get_axis(self):
        layout = self._canvas()._subplot_layout()
        return getattr(layout, self._axis)

    def _plt_get_tick_labels(self) -> tuple[list[float], list[str]]:
        texts = self._plt_get_axis().ticktext
        if texts is None:
            return [], []
        return self._plt_get_axis().tickvals, list(texts)

    def _plt_override_labels(self, pos: list[float], labels: list[str]):
        self._plt_get_axis().tickvals = pos
        self._plt_get_axis().ticktext = labels

    def _plt_reset_override(self):
        self._plt_get_axis().tickvals = None
        self._plt_get_axis().ticktext = None

    def _plt_get_visible(self) -> bool:
        return self._visible

    def _plt_set_visible(self, visible: bool):
        self._plt_get_axis().showticklabels = visible
        self._visible = visible

    def _plt_get_size(self) -> float:
        return self._plt_get_axis().tickfont.size

    def _plt_set_size(self, size: str):
        self._plt_get_axis().tickfont.size = size

    def _plt_get_fontfamily(self) -> str:
        return self._plt_get_axis().tickfont.family

    def _plt_set_fontfamily(self, font):
        self._plt_get_axis().tickfont.family = font

    def _plt_get_color(self):
        return np.array(self._plt_get_axis().tickfont.color)

    def _plt_set_color(self, color):
        self._plt_get_axis().tickfont.color = rgba_str_color(color)

    def _plt_get_text_rotation(self) -> float:
        return -self._plt_get_axis().tickangle

    def _plt_set_text_rotation(self, angle: float):
        self._plt_get_axis().tickangle = -angle


class Ticks3D(Ticks):
    def _plt_get_axis(self):
        layout = self._canvas()._subplot_layout()
        return getattr(layout.scene, self._axis)



from __future__ import annotations

from typing import Any

import numpy as np
from numpy.typing import NDArray

from whitecanvas.backend import Backend
from whitecanvas.layers._mixin import AbstractFaceEdgeMixin, MonoEdge, MultiFace
from whitecanvas.layers._primitive.line import LineLayerEvents
from whitecanvas.layers.layer3d._base import DataBoundLayer3D
from whitecanvas.protocols import MeshProtocol
from whitecanvas.types import ColorType, Hatch, LineStyle, MeshData


class MeshFace(MultiFace):
    def _ndata(self) -> int:
        return self._layer.data.faces.shape[0]


class Mesh3D(
    DataBoundLayer3D[MeshProtocol, MeshData], AbstractFaceEdgeMixin[MeshFace, MonoEdge]
):
    _backend_class_name = "components3d.Mesh3D"
    events: LineLayerEvents
    _events_class = LineLayerEvents

    def __init__(
        self,
        verts: NDArray[np.floating],
        faces: NDArray[np.intp],
        *,
        name: str | None = None,
        color: ColorType = "blue",
        hatch: Hatch = Hatch.SOLID,
        alpha: float = 1.0,
        backend: Backend | str | None = None,
    ):
        super().__init__(name=name)
        AbstractFaceEdgeMixin.__init__(self, MeshFace(self), MonoEdge(self))
        self._backend = self._create_backend(Backend(backend), verts, faces)
        self._x_hint, self._y_hint, self._z_hint = _verts_to_hints(verts)
        # self._backend._plt_connect_pick_event(self.events.clicked.emit)
        self.face.update(color=color, hatch=hatch, alpha=alpha)

    def _get_layer_data(self) -> MeshData:
        return MeshData(*self._backend._plt_get_data())

    def _norm_layer_data(self, data: Any) -> MeshData:
        verts, faces = data
        verts = np.atleast_2d(verts)
        faces = np.atleast_2d(faces)
        if verts.ndim != 2 or verts.shape[1] != 3:
            raise ValueError(f"Expected verts to be (N, 3), got {verts.shape}")
        if faces.ndim != 2 or faces.shape[1] != 3:
            raise ValueError(f"Expected faces to be (N, 3), got {faces.shape}")
        return MeshData(verts, faces)

    def _set_layer_data(self, data: MeshData):
        verts, faces = data
        self._backend._plt_set_data(verts, faces)
        self._x_hint, self._y_hint, self._z_hint = _verts_to_hints(verts)

    def set_data(self, verts: NDArray[np.floating], faces: NDArray[np.intp]):
        self.data = MeshData(verts, faces)

    def as_edge_only(
        self,
        *,
        width: float = 1.0,
        style: str | LineStyle = LineStyle.SOLID,
    ) -> Mesh3D:
        """
        Convert the surface to edge-only mode.

        This method will set the face color to transparent and the edge color to the
        current face color.

        Parameters
        ----------
        width : float, default 3.0
            Width of the edge.
        style : str or LineStyle, default LineStyle.SOLID
            Line style of the edge.
        """
        color = self.face.color
        if color.size == 0:
            ec = self.edge.color
        else:
            ec = np.mean(color, axis=0)
        self.with_edge(color=ec, width=width, style=style)
        self.face.update(alpha=0.0)
        return self


def _verts_to_hints(verts: NDArray[np.floating]):
    xmin, ymin, zmin = verts.min(axis=0)
    xmax, ymax, zmax = verts.max(axis=0)
    return (xmin, xmax), (ymin, ymax), (zmin, zmax)



import pandas as pd

from whitecanvas import new_canvas


def main():
    canvas = new_canvas(
        backend="matplotlib:qt",
        palette=["#EE6363", "#7777FF", "#57A557"],
    )

    url = "https://raw.githubusercontent.com/mwaskom/seaborn-data/master/iris.csv"
    df = pd.read_csv(url)
    cat_plt = canvas.cat_x(df, x="species", y="sepal_width")
    cat_plt.add_stripplot(
        color="species", extent=0.3
    ).with_edge(color="#3F3F00").move(-0.3)
    cat_plt.add_boxplot(color="species", extent=0.3)
    cat_plt.mean().add_markers(color="black", size=10, symbol="+")
    cat_plt.add_violinplot(
        color="species", extent=0.3, shape="right"
    ).with_edge(color="#3F3F00").move(0.2)

    canvas.show(block=True)

if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Generic,
    Sequence,
    TypeVar,
)

from whitecanvas import theme
from whitecanvas.canvas.dataframe._base import BaseCatPlotter
from whitecanvas.layers import tabular as _lt
from whitecanvas.layers.tabular import _jitter
from whitecanvas.types import HistBinType, KdeBandWidthType, Orientation

if TYPE_CHECKING:
    from typing_extensions import Self

    from whitecanvas.canvas._base import CanvasBase

    NStr = str | Sequence[str]

_C = TypeVar("_C", bound="CanvasBase")
_DF = TypeVar("_DF")


class _Aggregator(Generic[_C, _DF]):
    def __init__(self, method: str, plotter: CatPlotter[_C, _DF] = None):
        self._method = method
        self._plotter = plotter

    def __get__(self, ins: _C, owner) -> Self:
        return _Aggregator(self._method, ins)

    def __repr__(self) -> str:
        return f"Aggregator<{self._method}>"

    def __call__(self, by: str | tuple[str, ...]) -> CatPlotter[_C, _DF]:
        """Aggregate the values before plotting it."""
        plotter = self._plotter
        if isinstance(by, str):
            by = (by,)
        elif len(by) == 0:
            raise ValueError("No column is specified for grouping.")
        if plotter is None:
            raise TypeError("Cannot call this method from a class.")
        on = [plotter._x, plotter._y]
        df_agg = plotter._df.agg_by(by, on, self._method)
        return CatPlotter(
            plotter._canvas(),
            df_agg,
            x=plotter._x,
            y=plotter._y,
        )


class CatPlotter(BaseCatPlotter[_C, _DF]):
    """
    Categorical plotter that categorizes the data by features (color, style etc.)
    """

    def __init__(
        self,
        canvas: _C,
        df: _DF,
        x: str | None,
        y: str | None,
        update_labels: bool = False,
    ):
        super().__init__(canvas, df)
        self._x = x
        self._y = y
        self._update_label = update_labels
        if update_labels:
            self._update_xy_label(x, y)

    def _get_x(self) -> str:
        if self._x is None:
            raise ValueError("Column for x-axis is not set")
        return self._x

    def _get_y(self) -> str:
        if self._y is None:
            raise ValueError("Column for y-axis is not set")
        return self._y

    def _update_xy_label(self, x: str | None, y: str | None) -> None:
        """Update the x and y labels using the column names"""
        canvas = self._canvas()
        if isinstance(x, str):
            canvas.x.label.text = x
        if isinstance(y, str):
            canvas.y.label.text = y

    def along_x(self) -> CatPlotter[_C, _DF]:
        """Return the same plotter but with only x-axis set."""
        return self._copy_like(self._get_x(), None, self._update_label)

    def along_y(self) -> CatPlotter[_C, _DF]:
        """Return the same plotter but with only y-axis set."""
        return self._copy_like(None, self._get_y(), self._update_label)

    def _copy_like(self, x, y, update_label):
        out = self.__class__(self._canvas(), self._df, x, y, False)
        out._update_label = update_label
        return out

    def add_line(
        self,
        *,
        name: str | None = None,
        color: NStr | None = None,
        width: str | None = None,
        style: NStr | None = None,
    ) -> _lt.DFLines[_DF]:
        """
        Add a categorical line plot.

        >>> ### Use "time" column as x-axis and "value" column as y-axis
        >>> canvas.cat(df, "time", "value").add_line()

        >>> ### Multiple lines colored by column "group"
        >>> canvas.cat(df, "time", "value").add_line(color="group")

        >>> ### Multiple lines styled by column "group"
        >>> canvas.cat(df, "time", "value").add_line(style="group")

        Parameters
        ----------
        name : str, optional
            Name of the layer.
        color : str or sequence of str, optional
            Column name(s) for coloring the lines. Must be categorical.
        width : str, optional
            Column name for line width. Must be numerical.
        style : str or sequence of str, optional
            Column name(s) for styling the lines. Must be categorical.

        Returns
        -------
        DFLines
            Line collection layer.
        """
        canvas = self._canvas()
        width = theme._default("line.width", width)
        layer = _lt.DFLines.from_table(
            self._df, self._get_x(), self._get_y(), name=name, color=color, width=width,
            style=style, backend=canvas._get_backend(),
        )  # fmt: skip
        if color is not None and not layer._color_by.is_const():
            layer.update_color(layer._color_by.by, palette=canvas._color_palette)
        elif color is None:
            layer.update_color(canvas._color_palette.next())
        return canvas.add_layer(layer)

    def add_markers(
        self,
        *,
        name: str | None = None,
        color: NStr | None = None,
        hatch: NStr | None = None,
        size: str | None = None,
        symbol: NStr | None = None,
    ) -> _lt.DFMarkers[_DF]:
        """
        Add a categorical marker plot.

        >>> ### Use "time" column as x-axis and "value" column as y-axis
        >>> canvas.cat(df, "time", "value").add_markers()

        >>> ### Multiple markers colored by column "group"
        >>> canvas.cat(df, "time", "value").add_markers(color="group")

        >>> ### Change marker size according to "weight" column
        >>> canvas.cat(df, "time", "value").add_markers(size="weight")

        >>> ### Multiple markers with hatches determined by column "group"
        >>> canvas.cat(df, "time", "value").add_markers(hatch="group")

        >>> ### Multiple markers with symbols determined by "group"
        >>> canvas.cat(df, "time", "value").add_markers(symbol="group")

        Parameters
        ----------
        name : str, optional
            Name of the layer.
        color : str or sequence of str, optional
            Column name(s) for coloring the lines. Must be categorical.
        hatch : str or sequence of str, optional
            Column name(s) for hatches. Must be categorical.
        size : str, optional
            Column name for marker size. Must be numerical.
        symbol : str or sequence of str, optional
            Column name(s) for symbols. Must be categorical.

        Returns
        -------
        DFMarkers
            Marker collection layer.
        """
        canvas = self._canvas()
        xj = _jitter.IdentityJitter(self._get_x())
        yj = _jitter.IdentityJitter(self._get_y())
        layer = _lt.DFMarkers(
            self._df, xj, yj, name=name, color=color, hatch=hatch,
            size=size, symbol=symbol, backend=canvas._get_backend(),
        )  # fmt: skip
        if color is not None and not layer._color_by.is_const():
            layer.update_color(layer._color_by.by, palette=canvas._color_palette)
        elif color is None:
            layer.update_color(canvas._color_palette.next())
        return canvas.add_layer(layer)

    def add_hist2d(
        self,
        *,
        name: str | None = None,
        color: str | Sequence[str] | None = None,
        bins: HistBinType | tuple[HistBinType, HistBinType] = "auto",
        rangex: tuple[float, float] | None = None,
        rangey: tuple[float, float] | None = None,
        cmap=None,  # deprecated
    ) -> _lt.DFMultiHeatmap[_DF]:
        """
        Add 2-D histogram of given x/y columns.

        >>> ### Use "tip" column as x-axis and "total_bill" column as y-axis
        >>> canvas.cat(df, "tip", "total_bill").add_hist2d()

        Parameters
        ----------
        name : str, optional
            Name of the layer.
        color : str, optional
            Column name(s) for coloring the histogram.
        bins : int, array, str or tuple of them, default "auto"
            If int, the number of bins for both x and y. If tuple, the number of bins
            for x and y respectively.
        rangex : (float, float), optional
            Range of x values in which histogram will be built.
        rangey : (float, float), optional
            Range of y values in which histogram will be built.

        Returns
        -------
        DFMultiHeatmap
            Dataframe bound heatmap layer.
        """
        canvas = self._canvas()
        layer = _lt.DFMultiHeatmap.build_hist(
            self._df, self._get_x(), self._get_y(), color=color,name=name, bins=bins,
            range=(rangex, rangey), palette=canvas._color_palette,
            backend=canvas._get_backend(),
        )  # fmt: skip
        return canvas.add_layer(layer)

    def add_kde2d(
        self,
        *,
        name: str | None = None,
        color: str | None = None,
        band_width: KdeBandWidthType = "scott",
    ) -> _lt.DFMultiHeatmap[_DF]:
        """
        Add 2-D kernel density estimation of given x/y columns.

        >>> ### Use "tip" column as x-axis and "total_bill" column as y-axis
        >>> canvas.cat(df, "tip", "total_bill").add_kde2d()

        Parameters
        ----------
        name : str, optional
            Name of the layer.
        color : str, optional
            Column name(s) for coloring the densities.
        band_width : float, default None
            Bandwidth of the kernel density estimation. If None, use Scott's rule.

        Returns
        -------
        DFMultiHeatmap
            Dataframe bound heatmap layer.
        """
        canvas = self._canvas()
        layer = _lt.DFMultiHeatmap.build_kde(
            self._df, self._get_x(), self._get_y(), color=color, name=name,
            band_width=band_width, palette=canvas._color_palette,
            backend=canvas._get_backend(),
        )  # fmt: skip
        return canvas.add_layer(layer)

    def add_pointplot(
        self,
        *,
        name: str | None = None,
        color: NStr | None = None,
        hatch: NStr | None = None,
        size: float | None = None,
        capsize: float = 0.15,
    ) -> _lt.DFPointPlot2D[_DF]:
        """
        Add 2-D point plot.

        >>> ### Use "time" column as x-axis and "value" column as y-axis
        >>> canvas.cat(df, "time", "value").add_pointplot()

        >>> ### Multiple point plots colored by column "group"
        >>> canvas.cat(df, "time", "value").add_pointplot(color="group")

        >>> ### Multiple point plots with hatches determined by column "group"
        >>> canvas.cat(df, "time", "value").add_pointplot(hatch="group")

        Parameters
        ----------
        name : str, optional
            Name of the layer.
        color : str or sequence of str, optional
            Column name(s) for coloring the lines. Must be categorical.
        hatch : str or sequence of str, optional
            Column name(s) for hatches. Must be categorical.
        size : float, optional
            Size of the points.
        capsize : float, default 0.15
            Size of the cap on the error bars.

        Returns
        -------
        DFPointPlot2D
            Point plot layer.
        """
        canvas = self._canvas()
        layer = _lt.DFPointPlot2D(
            self._df, self._get_x(), self._get_y(), name=name, color=color,
            hatch=hatch, size=size, capsize=capsize, backend=canvas._get_backend(),
        )  # fmt: skip
        return canvas.add_layer(layer)

    def add_hist(
        self,
        *,
        bins: HistBinType = "auto",
        limits: tuple[float, float] | None = None,
        kind: str = "count",
        shape: str = "bars",
        name: str | None = None,
        color: NStr | None = None,
        width: float | None = None,
        style: NStr | None = None,
    ) -> _lt.DFHistograms[_DF]:
        """
        Add histograms.

        >>> ### Use "value" column as x-axis
        >>> canvas.cat(df, x="value").add_hist()

        >>> ### Multiple KDEs colored by column "group"
        >>> canvas.cat(df, x="value).add_hist(color="group")

        Parameters
        ----------
        bins : int, array-like or str, default "auto"
            If int, the number of bins. If array, the bin edges. If str, the method to
            calculate the number of bins.
        limits : (float, float), optional
            Lower and upper limits of the bins.
        kind : {"count", "density", "probability", "frequency"}, default "count"
            Kind of histogram to draw.
        shape : {"bars", "steps", "lines"}, default "bars"
            Shape of the histogram.
        name : str, optional
            Name of the layer.
        color : str or sequence of str, optional
            Column name(s) for coloring the lines. Must be categorical.
        width : str, optional
            Column name for line width. Must be numerical.
        style : str or sequence of str, optional
            Column name(s) for styling the lines. Must be categorical.

        Returns
        -------
        DFHistograms
            Histogram layer.
        """
        canvas = self._canvas()
        width = theme._default("line.width", width)
        x0, orient = self._column_and_orient()
        layer = _lt.DFHistograms.from_table(
            self._df, x0, bins=bins, limits=limits, kind=kind, shape=shape, name=name,
            orient=orient, color=color, width=width, style=style,
            backend=canvas._get_backend(),
        )  # fmt: skip
        if color is not None and not layer._color_by.is_const():
            layer.update_color(layer._color_by.by, palette=canvas._color_palette)
        elif color is None:
            layer.update_color(canvas._color_palette.next())
        if self._update_label:
            if orient.is_vertical:
                canvas.y.label.text = kind
            else:
                canvas.x.label.text = kind

        return canvas.add_layer(layer)

    def add_kde(
        self,
        *,
        band_width: KdeBandWidthType = "scott",
        name: str | None = None,
        color: NStr | None = None,
        width: str | None = None,
        style: NStr | None = None,
    ) -> _lt.DFKde[_DF]:
        """
        Add lines representing kernel density estimation.

        >>> ### Use "value" column as x-axis
        >>> canvas.cat(df, x="value").add_kde()

        >>> ### Multiple KDEs colored by column "group"
        >>> canvas.cat(df, x="value).add_kde(color="group")

        Parameters
        ----------
        band_width : float, default None
            Bandwidth of the kernel density estimation. If None, use Scott's rule.
        name : str, optional
            Name of the layer.
        color : str or sequence of str, optional
            Column name(s) for coloring the lines. Must be categorical.
        width : str, optional
            Column name for line width. Must be numerical.
        style : str or sequence of str, optional
            Column name(s) for styling the lines. Must be categorical.

        Returns
        -------
        DFKde
            KDE layer.
        """
        canvas = self._canvas()
        width = theme._default("line.width", width)
        x0, orient = self._column_and_orient()
        layer = _lt.DFKde.from_table(
            self._df, x0, band_width=band_width, name=name,
            orient=orient, color=color, width=width, style=style,
            backend=canvas._get_backend(),
        )  # fmt: skip
        if color is not None and not layer._color_by.is_const():
            layer.update_color(layer._color_by.by, palette=canvas._color_palette)
        elif color is None:
            layer.update_color(canvas._color_palette.next())
        if self._update_label:
            ax_label = "density"
            if orient.is_vertical:
                canvas.y.label.text = ax_label
            else:
                canvas.x.label.text = ax_label
        return canvas.add_layer(layer)

    def add_rug(
        self,
        *,
        name: str | None = None,
        color: NStr | None = None,
        width: NStr | float | None = None,
        style: NStr | None = None,
        low: float = 0.0,
        high: float | None = None,
    ) -> _lt.DFRug[_DF]:
        """
        Add a rug plot representing 1D distribution.

        Parameters
        ----------
        name : str, optional
            Name of the layer.
        color : str or sequence of str, optional
            Column name(s) for coloring the lines. Must be categorical.
        width : str, optional
            Column name for line width. Must be numerical.
        style : str or sequence of str, optional
            Column name(s) for styling the lines. Must be categorical.
        low : float, default 0.0
            Lower bound of each rug lines.
        high : float, default 0.0
            Higher bound of each rug lines. Automatically determined by the canvas size
            by default.

        Returns
        -------
        DFRug
            Rug plot layer.
        """
        canvas = self._canvas()
        width = theme._default("line.width", width)
        x0, orient = self._column_and_orient()
        if high is None:
            if orient.is_vertical:
                hmin, hmax = sorted(canvas.y.lim)
            else:
                hmin, hmax = sorted(canvas.x.lim)
            high = low + (hmax - hmin) * 0.05
        layer = _lt.DFRug.from_table(
            self._df, x0, name=name, orient=orient, color=color, width=width,
            style=style, low=low, high=high, backend=canvas._get_backend(),
        )  # fmt: skip
        if color is not None and not layer._color_by.is_const():
            layer.update_color(layer._color_by.by, palette=canvas._color_palette)
        elif color is None:
            layer.update_color(canvas._color_palette.next())
        return canvas.add_layer(layer)

    def _column_and_orient(self) -> tuple[str, Orientation]:
        if self._x is None and self._y is None:
            raise ValueError("Column for either x- or y-axis must be set")
        elif self._x is not None and self._y is not None:
            raise ValueError("Only one of x- or y-axis can be set")
        elif self._x is not None:
            return self._x, Orientation.VERTICAL
        else:
            return self._y, Orientation.HORIZONTAL

    # aggregators and group aggregators
    mean_for_each = _Aggregator("mean")
    median_for_each = _Aggregator("median")



from __future__ import annotations

import numpy as np
from numpy.typing import NDArray
from vispy.scene import visuals

from whitecanvas.backend import _not_implemented

# from vispy.visuals.filters.markers import MarkerPickingFilter
from whitecanvas.protocols import MarkersProtocol, check_protocol
from whitecanvas.types import Symbol
from whitecanvas.utils.normalize import as_color_array
from whitecanvas.utils.type_check import is_real_number


@check_protocol(MarkersProtocol)
class Markers(visuals.Markers):
    def __init__(self, xdata, ydata):
        pos = np.stack([xdata, ydata], axis=1)
        super().__init__(pos=pos, edge_width=0, face_color="blue")
        self.unfreeze()
        self._hover_texts: list[str] | None = None
        # self._picking_filter = MarkerPickingFilter()
        # self.attach(self._picking_filter)

    def _plt_get_ndata(self):
        return len(self._data["a_position"])

    ##### LayerProtocol #####
    def _plt_get_visible(self) -> bool:
        return self.visible

    def _plt_set_visible(self, visible: bool):
        self.visible = visible

    ##### XYDataProtocol #####
    def _plt_get_data(self):
        data = self._data["a_position"]
        return data[:, 0], data[:, 1]

    def _plt_set_data(self, xdata, ydata):
        ndata = self._plt_get_ndata()
        size = self._plt_get_symbol_size()
        edge_width = self._plt_get_edge_width()
        face_color = self._plt_get_face_color()
        edge_color = self._plt_get_edge_color()
        if xdata.size > ndata:
            size = np.concatenate([size, np.full(xdata.size - ndata, size[-1])])
            edge_width = np.concatenate(
                [edge_width, np.full(xdata.size - ndata, edge_width[-1])]
            )
            face_color = np.concatenate(
                [face_color, np.full((xdata.size - ndata, 4), face_color[-1])]
            )
            edge_color = np.concatenate(
                [edge_color, np.full((xdata.size - ndata, 4), edge_color[-1])]
            )
        elif xdata.size < ndata:
            size = size[: xdata.size]
            edge_width = edge_width[: xdata.size]
            face_color = face_color[: xdata.size]
            edge_color = edge_color[: xdata.size]
        self.set_data(
            pos=np.stack([xdata, ydata], axis=1),
            size=size,
            edge_width=edge_width,
            face_color=face_color,
            edge_color=edge_color,
            symbol=self.symbol[0],
        )

    ##### HasSymbol protocol #####
    def _plt_get_symbol(self) -> Symbol:
        if self._data["a_position"].shape[0] == 0:
            return Symbol.CIRCLE
        sym = self.symbol[0]
        if sym == "clobber":
            return Symbol.TRIANGLE_LEFT
        elif sym == "diamond":
            return Symbol.DIAMOND
        elif sym == "-":
            return Symbol.HBAR
        elif sym is None:
            return Symbol.CIRCLE
        return Symbol(sym)

    def _plt_set_symbol(self, symbol: Symbol):
        if symbol is Symbol.DIAMOND:
            self.symbol = "diamond"
        elif symbol is Symbol.HBAR:
            self.symbol = "-"
        elif symbol is Symbol.TRIANGLE_LEFT:
            # NOTE: vispy does not have "<"
            self.symbol = "clobber"
        else:
            self.symbol = symbol.value

    def _plt_get_symbol_size(self) -> NDArray[np.floating]:
        return self._data["a_size"]

    def _plt_set_symbol_size(self, size: float | NDArray[np.floating]):
        if is_real_number(size):
            size = np.full(self._plt_get_ndata(), size)
        if size.shape[0] == 0:
            return
        self.set_data(
            pos=self._data["a_position"],
            size=size,
            edge_width=self._plt_get_edge_width(),
            face_color=self._plt_get_face_color(),
            edge_color=self._plt_get_edge_color(),
            symbol=self.symbol,
        )

    ##### HasFace protocol #####
    def _plt_get_face_color(self) -> NDArray[np.float32]:
        return self._data["a_bg_color"]

    def _plt_set_face_color(self, color: NDArray[np.float32]):
        color = as_color_array(color, self._plt_get_ndata())
        if color.shape[0] == 0:
            return
        self.set_data(
            pos=self._data["a_position"],
            size=self._plt_get_symbol_size(),
            edge_width=self._plt_get_edge_width(),
            face_color=color,
            edge_color=self._plt_get_edge_color(),
            symbol=self.symbol,
        )

    _plt_get_face_hatch, _plt_set_face_hatch = _not_implemented.face_hatches()

    ##### HasEdges protocol #####
    def _plt_get_edge_color(self) -> NDArray[np.float32]:
        return self._data["a_fg_color"]

    def _plt_set_edge_color(self, color: NDArray[np.float32]):
        color = as_color_array(color, self._plt_get_ndata())
        if color.shape[0] == 0:
            return
        self.set_data(
            pos=self._data["a_position"],
            size=self._plt_get_symbol_size(),
            edge_width=self._plt_get_edge_width(),
            face_color=self._plt_get_face_color(),
            edge_color=color,
            symbol=self.symbol,
        )

    def _plt_get_edge_width(self) -> NDArray[np.floating]:
        return self._data["a_edgewidth"]

    def _plt_set_edge_width(self, width: float):
        if isinstance(width, float):
            width = np.full(self._plt_get_ndata(), width)
        if width.shape[0] == 0:
            return
        self.set_data(
            pos=self._data["a_position"],
            size=self._plt_get_symbol_size(),
            edge_width=width,
            face_color=self._plt_get_face_color(),
            edge_color=self._plt_get_edge_color(),
            symbol=self.symbol,
        )

    _plt_get_edge_style, _plt_set_edge_style = _not_implemented.edge_styles()

    def _plt_connect_pick_event(self, callback):
        # TODO: implement this
        # https://github.com/napari/napari/blob/main/napari/layers/points/points.py#L1617
        pass

    def _plt_set_hover_text(self, text: list[str]):
        # TODO: not used yet
        self._hover_texts = text

    def _compute_bounds(self, axis, view):
        # override to fix the bounds computation
        pos = self._data["a_position"]
        if pos is None or pos.size == 0:
            return None
        if pos.shape[1] > axis:
            return (pos[:, axis].min(), pos[:, axis].max())
        else:
            return (0, 0)



from __future__ import annotations

import numpy as np
from cmap import Colormap

from whitecanvas.layers._primitive import Image
from whitecanvas.layers.group._collections import LayerContainer
from whitecanvas.types import ColormapType, Orientation


class Colorbar(LayerContainer):
    _NO_PADDING_NEEDED = True

    def __init__(
        self,
        cmap: Colormap,
        *,
        name: str | None = None,
        orient: Orientation = Orientation.VERTICAL,
    ):
        self._cmap = cmap
        arr = _cmap_to_image(cmap, orient)
        image = Image(arr, name="lut")
        super().__init__([image], name=name)
        self._orient = orient

    @property
    def lut(self) -> Image:
        """The LUT image layer."""
        return self._children[0]

    @property
    def orient(self) -> Orientation:
        """The orientation of the colorbar."""
        return self._orient

    @property
    def shift(self):
        """The top-left corner of the colorbar."""
        return self.lut.shift

    @shift.setter
    def shift(self, value):
        self.lut.shift = value
        self._move_children()

    @property
    def scale(self):
        """The scale of the colorbar."""
        return self.lut.scale

    @scale.setter
    def scale(self, value):
        self.lut.scale = value
        self._move_children()

    def fit_to(self, bbox):
        self.lut.fit_to(bbox)
        self._move_children()

    @property
    def cmap(self) -> ColormapType:
        """The colormap of the colorbar."""
        return self._cmap

    @cmap.setter
    def cmap(self, cmap: ColormapType):
        """Set the colormap of the colorbar."""
        cmap = Colormap(cmap)
        arr = _cmap_to_image(cmap, self.orient)
        self.lut.data = arr
        self._cmap = cmap

    # def with_text(self, ) -> Colorbar:
    #     return self

    def _move_children(self):
        # Move other children to fit the colorbar.
        pass


def _cmap_to_image(cmap: Colormap, orient: Orientation):
    lut = cmap.lut()  # (N, 4)
    width = 50
    if orient.is_vertical:
        arr = np.repeat(lut[:, np.newaxis, :], width, axis=1)
    else:
        arr = np.repeat(lut[np.newaxis, :, :], width, axis=0)
    return arr



from __future__ import annotations

import weakref
from typing import TYPE_CHECKING, Generic, TypeVar

import numpy as np

from whitecanvas._exceptions import ReferenceDeletedError
from whitecanvas.layers import Layer, MultiLine
from whitecanvas.types import ColorType, LineStyle, XYData

if TYPE_CHECKING:
    from whitecanvas.canvas._base import CanvasBase

_C = TypeVar("_C", bound="CanvasBase")


class BetweenPlotter(Generic[_C]):
    def __init__(self, canvas: _C, layer1: Layer, layer2: Layer):
        self._canvas_ref = weakref.ref(canvas)
        self._layer1_ref = weakref.ref(layer1)
        self._layer2_ref = weakref.ref(layer2)

    def _canvas(self) -> _C:
        canvas = self._canvas_ref()
        if canvas is None:
            raise ReferenceDeletedError("Canvas has been deleted.")
        return canvas

    def _layers(self) -> tuple[Layer, Layer]:
        layer1 = self._layer1_ref()
        layer2 = self._layer2_ref()
        if layer1 is None or layer2 is None:
            raise ReferenceDeletedError("Layer has been deleted.")
        return layer1, layer2

    def connect_points(
        self,
        color: ColorType = "black",
        width: float = 1.0,
        alpha: float = 1.0,
        style: str | LineStyle = LineStyle.SOLID,
    ) -> MultiLine:
        canvas = self._canvas()
        layer1, layer2 = self._layers()
        if not isinstance(data1 := getattr(layer1, "data", None), XYData):
            raise TypeError(f"Layer {layer1!r} is not a layer with XY data.")
        if not isinstance(data2 := getattr(layer2, "data", None), XYData):
            raise TypeError(f"Layer {layer2!r} is not a layer with XY data.")
        if data1.x.size != data2.x.size:
            raise ValueError("Data sizes must be equal.")
        xs = np.stack([data1.x, data2.x], axis=1)
        ys = np.stack([data1.y, data2.y], axis=1)
        segs = np.stack([xs, ys], axis=2)
        lines = MultiLine(
            segs,
            name=f"between<{layer1.name}, {layer2.name}>",
            color=color,
            width=width,
            alpha=alpha,
            style=style,
        )
        canvas.add_layer(lines, under=[layer1, layer2])
        return lines

    # def add_bracket(
    #     self,
    #     string: str = "",
    #     capsize: float | None = None,
    #     *,
    #     color: ColorType = "black",
    #     width: float = 1.0,
    #     alpha: float = 1.0,
    #     style: str | LineStyle = LineStyle.SOLID,
    #     orientation: str | Orientation = Orientation.HORIZONTAL,
    # ):
    #     from whitecanvas.layers.group import BracketText

    #     canvas = self._canvas()
    #     layer1, layer2 = self._layers()
    #     hint1 = layer1.bbox_hint()
    #     hint2 = layer2.bbox_hint()

    #     ori = Orientation.parse(orientation)
    #     if ori.is_vertical:
    #         xmax = max(hint1[1], hint2[1])
    #         ymin = min(hint1[2], hint2[2])
    #         ymax = max(hint1[3], hint2[3])
    #         y1 = (hint1[2] + hint1[3]) / 2
    #         y2 = (hint2[2] + hint2[3]) / 2
    #         y1, y2 = sorted([y1, y2])
    #         capsize = 0.03 * (ymax - ymin)
    #         dy = capsize * 2
    #         layer = BracketText(
    #             (xmax, y2 + dy), (xmax, y1 + dy), string, capsize=capsize,
    #         )
    #     else:
    #         ymax = max(hint1[3], hint2[3])
    #         xmin = min(hint1[0], hint2[0])
    #         xmax = max(hint1[1], hint2[1])
    #         x1 = (hint1[0] + hint1[1]) / 2
    #         x2 = (hint2[0] + hint2[1]) / 2
    #         x1, x2 = sorted([x1, x2])
    #         capsize = 0.03 * (xmax - xmin)
    #         dx = capsize * 2
    #         layer = BracketText(
    #             (x1 + dx, ymax), (x2 + dx, ymax), string, capsize=capsize,
    #         )
    #     layer.line.update(color=color, width=width, alpha=alpha, style=style)
    #     return canvas.add_layer(layer)



from __future__ import annotations

import weakref
from typing import TYPE_CHECKING

import numpy as np
from cmap import Color
from qtpy import QtCore
from qtpy.QtGui import QFont, QPen

from whitecanvas.backend.pyqtgraph._base import PyQtAxis
from whitecanvas.backend.pyqtgraph._qt_utils import array_to_qcolor
from whitecanvas.types import LineStyle

if TYPE_CHECKING:
    import pyqtgraph as pg
    from qtpy import QtWidgets as QtW

    from whitecanvas.backend.pyqtgraph.canvas import Canvas

_X_AXIS = ("bottom", "top")


class _CanvasComponent:
    def __init__(self, canvas: Canvas):
        self._canvas = weakref.ref(canvas)


class Title(_CanvasComponent):
    def _plt_get_visible(self) -> bool:
        return self._canvas()._plot_item.titleLabel.isVisible()

    def _plt_set_visible(self, visible: bool):
        self._canvas()._plot_item.titleLabel.setVisible(visible)

    def _plt_get_text(self) -> str:
        return self._canvas()._plot_item.titleLabel.text

    def _plt_set_text(self, text: str):
        self._canvas()._plot_item.setTitle(text)

    def _plt_get_color(self):
        qcolor = self._canvas()._plot_item.titleLabel.opts["color"]
        return np.array(qcolor.getRgbF(), dtype=np.float32)

    def _plt_set_color(self, color):
        qcolor = array_to_qcolor(color)
        self._canvas()._plot_item.setTitle(self._plt_get_text(), color=qcolor)

    def _plt_get_size(self) -> int:
        pt = self._canvas()._plot_item.titleLabel.opts["size"]
        return int(pt[:-2])

    def _plt_set_size(self, size: int):
        self._canvas()._plot_item.setTitle(self._plt_get_text(), size=f"{size}pt")

    def _plt_get_fontfamily(self) -> str:
        return self._canvas()._plot_item.titleLabel.item.font().family()

    def _plt_set_fontfamily(self, family: str):
        self._canvas()._plot_item.titleLabel.item.setFont(
            QFont(family, self._plt_get_size())
        )


class AxisLabel(_CanvasComponent):
    def __init__(self, canvas: Canvas, axis: str):
        super().__init__(canvas)
        self._css = {
            "color": "#000000",
            "font-size": "11pt",
            "font-family": "Arial",
        }
        self._axis = axis

    def _get_axis(self) -> pg.AxisItem:
        return self._canvas()._plot_item.getAxis(self._axis)

    def _get_label(self) -> QtW.QGraphicsTextItem:
        return self._get_axis().label

    def _plt_get_visible(self) -> bool:
        return self._get_label().isVisible()

    def _plt_set_visible(self, visible: bool):
        self._get_axis().showLabel(visible)

    def _plt_get_text(self) -> str:
        return self._get_axis().labelText

    def _plt_set_text(self, text: str):
        self._get_axis().setLabel(text, **self._css)

    def _plt_get_color(self):
        return np.array(Color(self._css["color"]))

    def _plt_set_color(self, color):
        css = self._css.copy()
        css["color"] = Color(color).hex
        self._get_axis().setLabel(self._plt_get_text(), **css)
        self._css = css

    def _plt_get_size(self) -> int:
        return int(self._css["font-size"][:-2])

    def _plt_set_size(self, size: int):
        css = self._css.copy()
        css["font-size"] = f"{size}pt"
        self._get_axis().setLabel(self._plt_get_text(), **css)
        self._css = css

    def _plt_get_fontfamily(self) -> str:
        return self._css["font-family"]

    def _plt_set_fontfamily(self, family: str):
        css = self._css.copy()
        css["font-family"] = family
        self._get_axis().setLabel(self._plt_get_text(), **css)
        self._css = css


class Axis(_CanvasComponent):
    def __init__(self, canvas: Canvas, axis: str):
        super().__init__(canvas)
        self._axis = axis
        self._plt_get_axis().setZValue(-10)
        self._pen = QPen(array_to_qcolor(np.array([0.0, 0.0, 0.0, 1.0])))
        self._pen.setCosmetic(True)

    def _plt_get_axis(self) -> pg.AxisItem:
        return self._canvas()._plot_item.getAxis(self._axis)

    def _plt_get_limits(self) -> tuple[float, float]:
        return tuple(self._plt_get_axis().range)

    def _plt_set_limits(self, limits: tuple[float, float]):
        if self._axis in _X_AXIS:
            self._canvas()._viewbox().setXRange(*limits, padding=0)
        else:
            self._canvas()._viewbox().setYRange(*limits, padding=0)

    def _plt_get_color(self):
        return np.array(self._pen.color().getRgbF())

    def _plt_set_color(self, color):
        self._pen.setColor(array_to_qcolor(color))
        self._plt_get_axis().setTextPen(self._pen)
        self._plt_get_axis().setPen(self._pen)

    def _plt_flip(self) -> None:
        viewbox: pg.ViewBox = self._plt_get_axis().linkedView()
        if self._axis in _X_AXIS:
            viewbox.invertX()
        elif self._axis == "left":
            viewbox.invertY()

    def _plt_set_grid_state(self, visible: bool, color, width: float, style: LineStyle):
        if visible:
            grid = color[3] * 255
        else:
            grid = False
        axis = self._plt_get_axis()
        axis.setGrid(grid)
        # tick disappears by unknown reason.


class Ticks(_CanvasComponent):
    def __init__(self, canvas: Canvas, axis: str):
        super().__init__(canvas)
        self._axis = axis
        self._pen = QPen(array_to_qcolor(np.array([0.0, 0.0, 0.0, 1.0])))
        self._visible = True
        self._plt_get_axis().setTickFont(QFont("Arial"))  # avoid None

    def _plt_get_axis(self) -> pg.AxisItem:
        return self._canvas()._plot_item.getAxis(self._axis)

    def _plt_get_tick_labels(self) -> tuple[list[float], list[str]]:
        axis = self._plt_get_axis()
        if axis._tickLevels is not None:
            major: list[tuple[float, str]] = axis._tickLevels[0]
            values = []
            strings = []
            for val, string in major:
                values.append(val)
                strings.append(string)
        else:
            bounds = axis.mapRectFromParent(axis.geometry())
            span = (bounds.topLeft(), bounds.topRight())
            p0: QtCore.QPointF = axis.mapToDevice(span[0])
            p1: QtCore.QPointF = axis.mapToDevice(span[1])
            length_in_pixel = np.hypot(p0.x() - p1.x(), p0.y() - p1.y())
            tickvals = axis.tickValues(*axis.range, length_in_pixel)
            values = []
            for tickval in tickvals:
                values += list(tickval[1])
            values.sort()
            strings = values
        return values, strings

    def _plt_override_labels(self, pos: list[float], labels: list[str]):
        axis = self._plt_get_axis()
        axis.setTicks([list(zip(pos, labels))])

    def _plt_reset_override(self):
        self._plt_get_axis().setTicks(None)

    def _plt_get_visible(self) -> bool:
        return self._visible

    def _plt_set_fixed_size(self, size: float | None):
        if self._axis in _X_AXIS:
            self._plt_get_axis().setHeight(size)
        else:
            self._plt_get_axis().setWidth(size)

    def _plt_set_visible(self, visible: bool):
        axis = self._plt_get_axis()
        if visible:
            pen = self._pen
            self._plt_set_fixed_size(None)
        else:
            pen = QPen(array_to_qcolor(np.zeros(4)))
            self._plt_set_fixed_size(0)
        axis.setTickPen(pen)
        axis.setTextPen(pen)
        self._visible = visible

    def _get_font(self) -> QFont:
        return self._plt_get_axis().style["tickFont"]

    def _plt_get_size(self) -> float:
        return self._get_font().pointSizeF()

    def _plt_set_size(self, size: str):
        font = self._get_font()
        font.setPointSizeF(size)
        self._plt_get_axis().setTickFont(font)

    def _plt_get_fontfamily(self) -> str:
        return self._get_font().family()

    def _plt_set_fontfamily(self, font):
        qfont = self._get_font()
        qfont.setFamily(font)
        self._plt_get_axis().setTickFont(qfont)

    def _plt_get_color(self):
        return np.array(self._pen.color().getRgbF())

    def _plt_set_color(self, color):
        pen = self._pen
        pen.setColor(array_to_qcolor(color))
        self._plt_get_axis().setTickPen(pen)
        self._plt_get_axis().setTextPen(pen)

    def _plt_get_text_rotation(self) -> float:
        axis = self._plt_get_axis()
        if isinstance(axis, PyQtAxis):
            return axis.style["tickRotation"]
        return 0.0

    def _plt_set_text_rotation(self, rotation: float):
        axis = self._plt_get_axis()
        if isinstance(axis, PyQtAxis):
            axis.setTickRotation(rotation)



from typing import Any, Iterable, Literal, Sequence, Union

import numpy as np
from cmap import Color, Colormap
from numpy.typing import NDArray

ColorType = Union[str, Iterable["int | float"], Color]
ColormapType = Union[str, Colormap, "list[ColorType]", Any]
Number = Union[int, float, np.number]
ArrayLike1D = Union[Sequence[Number], NDArray[np.number]]
HistBinType = Union[int, ArrayLike1D, str]
KdeBandWidthType = Union[float, Literal["scott", "silverman"]]


class _Singleton:
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __repr__(self) -> str:
        cname = self.__class__.__name__.lstrip("_")
        return f"<{cname} object>"


class _Void(_Singleton):
    """Singleton that represents a void value."""



from functools import singledispatch

from matplotlib import patches
from matplotlib import transforms as mtransforms
from matplotlib.artist import Artist
from matplotlib.collections import LineCollection, PathCollection
from matplotlib.container import BarContainer, ErrorbarContainer, StemContainer
from matplotlib.lines import Line2D

from whitecanvas.backend.matplotlib._base import symbol_to_path
from whitecanvas.layers import _legend as _leg
from whitecanvas.types import Hatch


@singledispatch
def make_sample_item(item) -> "Artist | None":
    return None


@make_sample_item.register
def _(item: _leg.LineLegendItem):
    return Line2D(
        [], [], color=item.color, linewidth=item.width, linestyle=item.style.value
    )  # fmt: skip


@make_sample_item.register
def _(item: _leg.MarkersLegendItem):
    if item.face.hatch is Hatch.SOLID:
        hatch = None
    else:
        hatch = item.face.hatch.value
    col = PathCollection(
        (symbol_to_path(item.symbol),),
        sizes=[item.size**2],
        offsets=[(0, 0)],
        facecolors=item.face.color,
        edgecolors=item.edge.color,
        linewidths=[item.edge.width],
        linestyles=[item.edge.style.value],
        hatch=hatch,
    )
    col.set_transform(mtransforms.IdentityTransform())
    return col


@make_sample_item.register
def _(item: _leg.BarLegendItem):
    rect = patches.Rectangle(
        (0, 0), 1, 1, facecolor=item.face.color, edgecolor=item.edge.color,
        linewidth=item.edge.width, linestyle=item.edge.style.value,
        hatch=item.face.hatch.value * 4
    )  # fmt: skip
    return BarContainer([rect])


@make_sample_item.register
def _(item: _leg.PlotLegendItem):
    return Line2D(
        [], [], color=item.line.color, linewidth=item.line.width,
        linestyle=item.line.style.value, marker=item.markers.symbol.value,
        markersize=item.markers.size, markerfacecolor=item.markers.face.color,
        markeredgecolor=item.markers.edge.color,
        markeredgewidth=item.markers.edge.width,
    )  # fmt: skip


@make_sample_item.register
def _(item: _leg.ErrorLegendItem):
    line_2d = Line2D([], [], color="#00000000")
    caps = Line2D([], [], color=item.color, linewidth=item.width)
    barlines = LineCollection(
        [], color=item.color, linewidth=item.width, linestyle=item.style.value
    )
    return ErrorbarContainer((line_2d, (caps,), [barlines]), has_yerr=True)


@make_sample_item.register
def _(item: _leg.LineErrorLegendItem):
    line_2d = make_sample_item(item.line)
    caps, barlines, has_xerr, has_yerr = _norm_xyerr(item.xerr, item.yerr)
    return ErrorbarContainer(
        (line_2d, caps, barlines), has_xerr=has_xerr, has_yerr=has_yerr
    )


@make_sample_item.register
def _(item: _leg.MarkerErrorLegendItem):
    marker = _marker_as_line_2d(item.markers)
    caps, barlines, has_xerr, has_yerr = _norm_xyerr(item.xerr, item.yerr)
    return ErrorbarContainer(
        (marker, caps, barlines), has_xerr=has_xerr, has_yerr=has_yerr
    )


@make_sample_item.register
def _(item: _leg.PlotErrorLegendItem):
    plot = make_sample_item(item.plot)
    caps, barlines, has_xerr, has_yerr = _norm_xyerr(item.xerr, item.yerr)
    return ErrorbarContainer(
        (plot, caps, barlines), has_xerr=has_xerr, has_yerr=has_yerr
    )


@make_sample_item.register
def _(item: _leg.StemLegendItem):
    markers = _marker_as_line_2d(item.markers)
    markers.set_markersize(item.markers.size / 1.6)  # to make stem visible
    stemlines = make_sample_item(item.line)
    baseline = Line2D([], [], color="#00000000")
    return StemContainer((markers, [stemlines], baseline))


@make_sample_item.register
def _(item: _leg.TitleItem):
    return patches.Rectangle((0, 0), 0, 0, color="#00000000")


def _norm_xyerr(xerr: "_leg.LineLegendItem | None", yerr: "_leg.LineLegendItem | None"):
    has_xerr = xerr is not None
    has_yerr = yerr is not None
    if has_xerr:
        er = _make_line_from_err(xerr)
        barline = _make_line_collection_from_err(xerr)
    elif has_yerr:
        er = _make_line_from_err(yerr)
        barline = _make_line_collection_from_err(yerr)
    else:
        er = None
        barline = []
    if er is None:
        caps = ()
    else:
        caps = (er,)
    return caps, [barline], has_xerr, has_yerr


def _make_line_from_err(item: _leg.LineLegendItem):
    return Line2D(
        [], [], color=item.color, linewidth=item.width, linestyle=item.style.value
    )  # fmt: skip


def _make_line_collection_from_err(item: _leg.LineLegendItem):
    return LineCollection(
        [], color=item.color, linewidth=item.width, linestyle=item.style.value
    )  # fmt: skip


def _marker_as_line_2d(item: _leg.MarkersLegendItem):
    return Line2D(
        [], [], color=None, marker=item.symbol.value, markersize=item.size,
        markerfacecolor=item.face.color, markeredgecolor=item.edge.color,
        markeredgewidth=item.edge.width
    )  # fmt: skip
